"""Plot store."""
import typing as ty

import numpy as np
from bokeh.models import ColumnDataSource

from plotski.spectrum.custom import (
    PlotButterflyMassSpectrum,
    PlotButterflyMobilogram,
    PlotCentroidMassSpectrum,
    PlotMassSpectrum,
    PlotMobilogram,
)
from plotski.store import PlotStore


class CustomPlotStore(PlotStore):
    """Plot store with extra functionality."""

    def plot_mass_spectrum(self, tab_name, data: ty.Dict, layout_name=None, **kwargs):
        """Adds mass spectrum to the plot store

        Parameters
        ----------
        tab_name : str
            name of the tab where plot should be added to
        data : dict
            dictionary containing appropriate plot fields
            in this case:
                x = list / array
                y = list / array
            the length of x and y must be the same
        layout_name : str
            by default, plot objects are added to the tab in iterative way (e.g. if there are no plots in the tab, it
            will be added as 'item #0', if there is one then it will be added as 'item #1' etc. Sometimes you might want
            to add it to a 'row' or 'column' for which you have name - you can specify its name here and if its present
            the plot object will be added to that container
        kwargs :
            dictionary containing plot parameters e.g. x/y axis labels, title, etc...

        Returns
        -------
        tab_name : str
            name of the tab
        item_name : str
            name of the plot
        plot : PlotMassSpectrum
            plot object
        """
        self.check_tab(tab_name)
        self.check_data(data, ("x", "y"))

        source = ColumnDataSource(data)
        plot = PlotMassSpectrum(self.output_dir, source=source, **kwargs)

        # add figure object to tab
        layout_name = layout_name if layout_name is not None else self.get_unique_name(tab_name)
        plot = self.append_item(tab_name, layout_name, plot)
        return tab_name, layout_name, plot

    def plot_butterfly_mass_spectrum(self, tab_name, data: ty.Dict, layout_name=None, **kwargs):
        """Adds butterfly mass spectra to the plot store (one on top / one below)

        Parameters
        ----------
        tab_name : str
            name of the tab where plot should be added to
        data : dict
            Dictionary containing appropriate plot fields, in this case:
                x_top, y_top, x_bottom, y_bottom = list / array
            the length of x_top and y_top, x_bottom and y_bottom must be the same.
        layout_name : str
            by default, plot objects are added to the tab in iterative way (e.g. if there are no plots in the tab, it
            will be added as 'item #0', if there is one then it will be added as 'item #1' etc. Sometimes you might want
            to add it to a 'row' or 'column' for which you have name - you can specify its name here and if its present
            the plot object will be added to that container
        kwargs :
            dictionary containing plot parameters e.g. x/y axis labels, title, etc...

        Returns
        -------
        tab_name : str
            name of the tab
        item_name : str
            name of the plot
        plot : PlotButterflyMassSpectrum
            plot object
        """
        self.check_tab(tab_name)
        self.check_data(data, ("x_top", "y_top", "x_bottom", "y_bottom"))

        source = ColumnDataSource(data=data)
        plot = PlotButterflyMassSpectrum(self.output_dir, source=source, **kwargs)

        # add figure object to tab
        layout_name = layout_name if layout_name is not None else self.get_unique_name(tab_name)
        self.append_item(tab_name, layout_name, plot)
        return tab_name, layout_name, plot

    def plot_centroid_mass_spectrum(self, tab_name, data: ty.Dict, layout_name=None, **kwargs):
        """Adds centroid mass spectrum to the plot store

        Parameters
        ----------
        tab_name : str
            name of the tab where plot should be added to
        data : dict
            Dictionary containing appropriate plot fields, in this case:
                x, y0, y1 = list / array
            the length of x and y0 and y1 must be the same.
            If y0 is not provided, values will be automatically generated to provide array of 0s
        layout_name : str
            by default, plot objects are added to the tab in iterative way (e.g. if there are no plots in the tab, it
            will be added as 'item #0', if there is one then it will be added as 'item #1' etc. Sometimes you might want
            to add it to a 'row' or 'column' for which you have name - you can specify its name here and if its present
            the plot object will be added to that container
        kwargs :
            dictionary containing plot parameters e.g. x/y axis labels, title, etc...

        Returns
        -------
        tab_name : str
            name of the tab
        item_name : str
            name of the plot
        plot : PlotCentroidMassSpectrum
            plot object
        """
        self.check_tab(tab_name)
        if "y0" not in data:
            data["y0"] = np.zeros_like(data["x"], dtype=np.int8)
        self.check_data(data, ("x", "y0", "y1"))

        source = ColumnDataSource(data)
        plot = PlotCentroidMassSpectrum(self.output_dir, source=source, **kwargs)

        # add figure object to tab
        layout_name = layout_name if layout_name is not None else self.get_unique_name(tab_name)
        self.append_item(tab_name, layout_name, plot)
        return tab_name, layout_name, plot

    def plot_mobilogram(self, tab_name, data: ty.Dict, layout_name=None, **kwargs):
        """Adds mobilogram to the plot store

        Parameters
        ----------
        tab_name : str
            name of the tab where plot should be added to
        data : dict
            dictionary containing appropriate plot fields
            in this case:
                x, y = list / array
            the length of x and y must be the same
        layout_name : str
            by default, plot objects are added to the tab in iterative way (e.g. if there are no plots in the tab, it
            will be added as 'item #0', if there is one then it will be added as 'item #1' etc. Sometimes you might want
            to add it to a 'row' or 'column' for which you have name - you can specify its name here and if its present
            the plot object will be added to that container
        kwargs :
            dictionary containing plot parameters e.g. x/y axis labels, title, etc...

        Returns
        -------
        tab_name : str
            name of the tab
        item_name : str
            name of the plot
        plot : PlotMobilogram
            plot object
        """
        self.check_tab(tab_name)
        self.check_data(data, ("x", "y"))

        source = ColumnDataSource(data)
        plot = PlotMobilogram(self.output_dir, source=source, **kwargs)

        # add figure object to tab
        layout_name = layout_name if layout_name is not None else self.get_unique_name(tab_name)
        self.append_item(tab_name, layout_name, plot)
        return tab_name, layout_name, plot

    def plot_butterfly_mobilogram(self, tab_name, data: ty.Dict, layout_name=None, **kwargs):
        """Adds butterfly mobilograms to the plot store (one on top / one below)

        Parameters
        ----------
        tab_name : str
            name of the tab where plot should be added to
        data : dict
            Dictionary containing appropriate plot fields, in this case:
                x_top, y_top, x_bottom, y_bottom = list / array
            the length of x_top and y_top, x_bottom and y_bottom must be the same.
        layout_name : str
            by default, plot objects are added to the tab in iterative way (e.g. if there are no plots in the tab, it
            will be added as 'item #0', if there is one then it will be added as 'item #1' etc. Sometimes you might want
            to add it to a 'row' or 'column' for which you have name - you can specify its name here and if its present
            the plot object will be added to that container
        kwargs :
            dictionary containing plot parameters e.g. x/y axis labels, title, etc...

        Returns
        -------
        tab_name : str
            name of the tab
        item_name : str
            name of the plot
        plot : PlotButterflyMobilogram
            plot object
        """
        self.check_tab(tab_name)
        self.check_data(data, ("x_top", "y_top", "x_bottom", "y_bottom"))

        source = ColumnDataSource(data)
        plot = PlotButterflyMobilogram(self.output_dir, source=source, **kwargs)

        # add figure object to tab
        layout_name = layout_name if layout_name is not None else self.get_unique_name(tab_name)
        self.append_item(tab_name, layout_name, plot)
        return tab_name, layout_name, plot



"""Module containing functions that generate RGB plots"""
import typing as ty
from itertools import cycle

import numpy as np
from matplotlib.colors import ListedColormap
from skimage import exposure
from skimage.color import gray2rgb, gray2rgba

from plotski.utilities import convert_hex_to_rgb_1, rescale

np.seterr(divide="ignore", invalid="ignore")

COLORS = cycle(
    [
        "#800000",
        "#9A6324",
        "#469990",
        "#000075",
        "#E6194B",
        "#3CB44B",
        "#A9A9A9",
        "#F58231",
        "#FFC119",
        "#3CB44B",
        "#42D4F4",
        "#F032E6",
        "#AAFFC3",
    ]
)


class ImageRGBA:
    """This class simplifies generation of composite images by remapping individual ion images to their respective
    colors, which can be quite useful when visualising multiple species at the same time.

    Parameters
    ----------
    images : List[np.ndarray]
        list of flat images
    colors : Optional[List]
        list of colors - if None have been specified, a set of defaults will be used instead
    """

    def __init__(self, images: ty.List[np.ndarray], colors: ty.Optional[ty.List] = None):
        """Class to quickly generate composite RGBA images based on ion (or other) images"""
        self.validate(images, colors)

        self._original, self._images, self._colors = self.setup(images, colors)
        self._rgba: ty.Optional[np.ndarray] = None

    def __repr__(self):
        return f"ImageRGBA <images={len(self._images)}>"

    @staticmethod
    def _convert_color(color):
        """Convert hex/rgb color"""
        if isinstance(color, str):
            return convert_hex_to_rgb_1(color)
        if isinstance(color, (tuple, list, np.ndarray)):
            if np.max(color) > 1.0:
                color = np.asarray(color) / 255
            return color

    @property
    def intensities(self) -> np.ndarray:
        """Return summed intensities from the original data"""
        return np.sum(self._original, axis=0)

    @property
    def rgb(self) -> np.ndarray:
        """Return combined image array without the alpha channel"""
        return self.rgba[:, :, 0:3].astype(np.uint8)

    @property
    def rgba(self) -> np.ndarray:
        """Return combined image array with the alpha channel"""
        if self._rgba is None:
            self._rgba = self.combine().astype(np.uint8)
            self._rgba[:, :, 3] = 255
        return self._rgba

    @rgba.setter
    def rgba(self, array: ty.Optional[np.ndarray]):
        if array is not None and array.shape[2] != 4:
            raise ValueError("Cannot set RGBA array without the alpha channel")
        self._rgba = array

    @staticmethod
    def validate(images: ty.List[np.ndarray], colors: ty.Optional[ty.List[str]]) -> None:
        """Validate image input

        Parameters
        ----------
        images : List[np.ndarray]
            list of 2D numpy arrays
        colors : Optional[List[str]]
            list of hex colors

        Raises
        ------
        ValueError
            raised if image shapes are not compatible
        """
        if not isinstance(images, list):
            raise ValueError("Expected list of images")

        images = [np.asarray(image) for image in images]
        shapes = set()
        for image in images:
            shapes.add(image.shape)

        if len(shapes) > 1:
            raise ValueError("Images must be of the same size")

        # if color list was specified, make sure the values are correct
        if colors is not None:
            if len(images) != len(colors):
                raise ValueError(
                    f"Please provide the same number of colors as images. Number of images: {len(images)}"
                    f" and colors: {len(colors)}"
                )
            for color in colors:
                if isinstance(color, str):
                    if not color.startswith("#"):
                        raise ValueError("Hex color must start with a hash e.g. `#FF00FF`")
                elif isinstance(color, (list, tuple, np.ndarray)):
                    color_len = len(color)
                    if color_len not in [3, 4]:
                        raise ValueError("RGB(A) color must have between 3-4 values (R, G, B and A)")
                else:
                    raise ValueError("Color must be a string (hex) or list (RGBA in range 0-1)")

    def reset(self):
        """Reset the store RGBA array to ignore previously made changes (e.g. channel normalizations)"""
        self.rgba = None

    def setup(
        self, images: ty.List[np.ndarray], colors: ty.Optional[ty.List[str]]
    ) -> ty.Tuple[ty.List, ty.List, ty.List]:
        """Clean-up and remap images to appropriate color

        Parameters
        ----------
        images : List[np.ndarray]
            list of 2D numpy arrays
        colors : Optional[List[str]]
            list of hex colors

        Returns
        -------
        images : List
            list of remapped images
        _images : List
            list of original images
        colors : Iterable
            color iterable
        """
        if colors is None:
            colors = [next(COLORS) for _ in images]

        _images = []
        for image, color in zip(images, colors):
            if len(image.shape) == 2:
                image = self._to_rgb(image, self._convert_color(color))
            _images.append(image)

        return images, _images, colors

    def rescale(self, channel_id: int, max_value: ty.Union[int, float] = 255):
        """Rescale array to standardized range."""
        return rescale(np.nan_to_num(self._original[channel_id]), 0, max_value)

    @staticmethod
    def _to_rgb(
        image: np.ndarray,
        color: ty.Union[ty.List, ty.Tuple],
        add_alpha: bool = True,
        max_value: ty.Union[int, float] = 255,
    ) -> np.ndarray:
        """Convert flat array to multi-channel RGB image based on the specified color

        Parameters
        ----------
        image : np.ndarray
            flat image array
        color : Union[Tuple, List]
            chosen RGB color in the scale 0-255
        add_alpha : bool
            if `True`, 4-th alpha channel will be added to the image array
        max_value : Union[int, float]
            maximum value the image should be rescaled to

        Returns
        -------
        rgb : np.ndarray
            3/4-dimensional image array remapped to the `color`

        """
        if add_alpha:
            image = gray2rgba(rescale(np.nan_to_num(image), 0, max_value))
        else:
            image = gray2rgb(rescale(np.nan_to_num(image), 0, max_value))

        if add_alpha and len(color) == 3:
            if isinstance(color, np.ndarray):
                np.append(color, [1.0])
            else:
                color = list(color)
                color.append(1.0)

        return image * color

    def recolor(self, image_id: int, color: ty.Union[str, ty.List]):
        """Change color of particular image

        Parameters
        ----------
        image_id : int
            index of the image
        color : Union[str, List]
            new color of the image
        """
        if image_id > len(self._images) - 1:
            raise ValueError("Cannot update color of image that is not present")

        self._colors[image_id] = color
        self._images[image_id] = self._to_rgb(self._original[image_id], self._convert_color(color))
        self.reset()

    def get_one(self, image_id: int, keep_alpha: bool = False, dtype=np.uint8, fill_alpha: ty.Optional[int] = None):
        """Retrieve single 3/4D image

        Parameters
        ----------
        image_id : int
            index of the image
        keep_alpha : bool, optional
            if `True`, the alpha channel will be retained in the image
        dtype :
            data type of the image - matplotlib and Bokeh expect uint8 values to represent images
        fill_alpha : int, optional
            if value is an integer and `keep_alpha` is True, alpha channel will be filled with the specified value

        Returns
        -------
        array : np.ndarray
            3/4D image array mapped to specific color
        """
        if image_id > len(self._images) - 1:
            raise ValueError("Tried to retrieve image that is not present")

        if keep_alpha:
            image = self._images[image_id].astype(dtype)
            if fill_alpha is not None and isinstance(fill_alpha, int):
                image[:, :, 3] = fill_alpha
            return image
        return self._images[image_id][:, :, :3].astype(dtype)

    def get_rgba(self, fill_alpha: ty.Optional[int] = None):
        """Retrieve RGBA image"""
        image = self.rgba
        if fill_alpha is not None and isinstance(fill_alpha, int):
            image[:, :, 3] = fill_alpha

        return image

    def combine(self, max_value: ty.Union[int, float] = 255, dtype=np.uint8) -> np.ndarray:
        """Combine multiple images into one

        Parameters
        ----------
        max_value : Union[int, float]
            maximum value the image should be rescaled to
        dtype : np.dtype
            numpy data type

        Returns
        -------
        combined_rgb : nd.array
            3/4-dimensional image from multiple images
        """
        return self._combine(self._images, max_value, dtype)

    @staticmethod
    def _combine(images: ty.List[np.ndarray], max_value: ty.Union[int, float], dtype) -> np.ndarray:
        """Combine multiple images into one

        Parameters
        ----------
        images : List[np.ndarray]
            list of RGBA images
        max_value : Union[int, float]
            maximum value the image should be rescaled to
        dtype : np.dtype
            numpy data type

        Returns
        -------
        combined_rgb : nd.array
            3/4-dimensional image from multiple images
        """
        combined_rgb = np.sum(images, axis=0)
        combined_rgb = np.clip(combined_rgb, 0, max_value)
        return combined_rgb.astype(dtype)

    def clip_channel(
        self,
        channel: int,
        min_int: ty.Optional[ty.Union[int, float]] = None,
        max_int: ty.Optional[ty.Union[int, float]] = None,
    ):
        """Normalize image channel to increase/decrease intensity

        Parameters
        ----------
        channel : int
            channel to normalize
        min_int : Optional[Union[int, float]]
            scalar value to normalize the minimum of the channel to
        max_int : Optional[Union[int, float]]
            scalar value to normalize the maximum of the channel to

        Returns
        -------
        self : ImageRGBA
            returns self to enable chaining of multiple commands in one go

        Examples
        --------
        Instantiate ImageRGBA and normalize each channel to the same intensity range between 20-200

        >>> np.random.seed(42)
        >>> images = [np.random.randint(0, 255, (3, 3))]
        >>> image_rgba = ImageRGBA(images, colors=["#800000"])
        >>> image_rgba.clip_channel(0, 20, 200).clip_channel(1, 20, 200).clip_channel(2, 20, 200).rgba
        array([[[ 64, 200, 200, 255],
                [120, 200, 200, 255],
                [ 57, 200, 200, 255]],
        <BLANKLINE>
               [[ 20, 200, 200, 255],
                [ 67, 200, 200, 255],
                [ 41, 200, 200, 255]],
        <BLANKLINE>
               [[200, 200, 200, 255],
                [ 20, 200, 200, 255],
                [ 64, 200, 200, 255]]], dtype=uint8)
        """
        # if image is None:
        image = self.rgba
        v_min = image[:, :, channel].min()
        v_max = image[:, :, channel].max()
        self.rgba = self._clip_channel(image, channel, v_min, v_max, min_int, max_int)
        return self

    @staticmethod
    def _clip_channel(
        image: np.ndarray,
        channel: int,
        v_min: ty.Union[int, float],
        v_max: ty.Union[int, float],
        min_value: ty.Union[int, float] = 0,
        max_value: ty.Union[int, float] = 255,
    ):
        """Normalize image channel to increase/decrease intensity

        Parameters
        ----------
        image : np.ndarray
            3/4D image array
        channel : int
            channel to normalize
        v_min : Union[int, float]
            scalar value to normalize the minimum of the channel to
        v_max : Union[int, float]
            scalar value to normalize the maximum of the channel to
        min_value : Union[int, float]
            minimum clipping value
        max_value : Union[int, float]
            maximum clipping value
        Returns
        -------
        image : np.ndarray
            3/4D image with one of the channels normalized to the `vmin`/`vmax` values
        """
        im_channel = image[:, :, channel]
        im_channel[im_channel <= v_min] = min_value
        im_channel[im_channel >= v_max] = max_value
        im_channel = np.clip(im_channel, min_value, max_value)
        image[:, :, channel] = im_channel
        return image

    # noinspection PyTypeChecker
    def quantile_rescale(self, q_low: float = 0.02, q_high: float = 0.98, image: ty.Optional[np.ndarray] = None):
        """Contrast enhancement using stretching or shrinking of intensity levels

        Returns
        -------
        q_low : float
            low boundary quantile
        q_high : float
            high boundary quantile
        image : Optional[np.ndarray]
            image array, if one is not provided, the `rgba` attribute will be used instead

        Returns
        -------
        rgb_image : np.ndarray
            image after rescaling
        """
        if 0 < q_low > 1:
            raise ValueError(f"`q_low` should be between 0 and 1 (not {q_low})")
        if 0 < q_high > 1:
            raise ValueError(f"`q_high` should be between 0 and 1 (not {q_high})")
        if image is None:
            image = self.rgba
        p_low, p_high = np.quantile(image, [q_low, q_high])
        return exposure.rescale_intensity(image, in_range=(p_low, p_high))

    # noinspection PyTypeChecker
    def equalize_histogram(self, n_bins: int = 256, image: ty.Optional[np.ndarray] = None, as_int: bool = False):
        """Contrast enhancement using histogram equalization

        Parameters
        ----------
        n_bins : int, optional
            number of bins for image histogram
        image : Optional[np.ndarray]
            image array, if one is not provided, the `rgba` attribute will be used instead
        as_int : bool, optional
            if ``True`` the returned array will be multiplied by 255 and converted to uint8

        Returns
        -------
        rgb_image : np.ndarray
            equalized image
        """
        if image is None:
            image = self.rgba
        array = exposure.equalize_hist(image, n_bins)
        if as_int:
            return (array * 255).astype(np.uint8)
        return array

    def adaptive_histogram(
        self, clip_limit: float = 0.01, n_bins: int = 256, image: ty.Optional[np.ndarray] = None, as_int: bool = False
    ):
        """Contrast Limited Adaptive Histogram Equalization

        Locally enhance contrast of a RGBA image. The algorithm computes histograms over different tile regions of
        the image.

        Parameters
        ----------
        clip_limit : float, optional
            image clipping limit, normalized between 0 and 1 (higher values give more contrast)
        n_bins : int, optional
            number of gray bins for histogram
        image : Optional[np.ndarray]
            image array, if one is not provided, the `rgba` attribute will be used instead
        as_int : bool, optional
            if ``True`` the returned array will be multiplied by 255 and converted to uint8

        Returns
        -------
        rgb_image : np.ndarray
            equalized image
        """
        if image is None:
            image = self.rgba
        array = exposure.equalize_adapthist(image, nbins=n_bins, clip_limit=clip_limit)
        if as_int:
            return (array * 255).astype(np.uint8)
        return array

    def contrast_stretching(self, in_range="image", image: ty.Optional[np.ndarray] = None, as_int: bool = False):
        """Contrast Limited Adaptive Histogram Equalization

        Locally enhance contrast of a RGBA image. The algorithm computes histograms over different tile regions of
        the image.

        Parameters
        ----------
        in_range: str or 2-tuple
            Min and max intensity values of the input image. Possible values include: `image`, `dtype`, `dtype-name` or
            a two-element tuple including two quantiles
        image : Optional[np.ndarray]
            image array, if one is not provided, the `rgba` attribute will be used instead
        as_int : bool, optional
            if ``True`` the returned array will be multiplied by 255 and converted to uint8

        Returns
        -------
        rgb_image : np.ndarray
            equalized image
        """
        if image is None:
            image = self.rgba
        if isinstance(in_range, tuple):
            q1, q2 = np.quantile(image, in_range)
            in_range = (q1, q2)
        array = exposure.rescale_intensity(image, in_range)
        if as_int:
            return (array * 255).astype(np.uint8)
        return array

    def get_colormap(self, image_id: int, n_bins: int = 256, name: str = "colormap"):
        """Create linear colormap

        Parameters
        ----------
        image_id : int
            index of the image/color
        n_bins : int, optional
            number of values in the colormap
        name : str, optional
            name of the colormap

        Returns
        -------
        colormap : ListedColormap
            colormap
        """
        if image_id > len(self._images) - 1:
            raise ValueError("Tried to retrieve image that is not present")

        color = self._convert_color(self._colors[image_id])
        lin_range = np.linspace(0, 1.0, n_bins)
        array = np.zeros((n_bins, len(color)))
        array[:, 0] = lin_range * color[0]
        array[:, 1] = lin_range * color[1]
        array[:, 2] = lin_range * color[2]

        colormap = ListedColormap(array, name)
        return colormap



# Third-party imports
import numpy as np
import pytest
from matplotlib.colors import ListedColormap
from numpy.testing import assert_array_equal

# Local imports
from plotski.rgb import ImageRGBA
from plotski.utilities import get_random_hex_color


def get_images(n_items):
    return [np.random.randint(0, 255, (10, 10)) for _ in range(n_items)]


def get_colors(n_items):
    return [get_random_hex_color() for _ in range(n_items)]


class TestImageRGBA:
    @staticmethod
    @pytest.mark.parametrize("n_images", (1, 3, 5))
    def test_init_images_hex(n_images):
        images = get_images(n_images)
        image_rgb = ImageRGBA(images)
        assert image_rgb
        assert len(image_rgb._original) == len(image_rgb._images) == len(image_rgb._colors) == n_images

        assert "images=" in repr(image_rgb)

    @staticmethod
    @pytest.mark.parametrize("n_images", (1, 3, 5))
    def test_init_images_rgb(n_images):
        images = get_images(n_images)
        image_rgb = ImageRGBA(images, colors=[[1, 0, 0.7]] * n_images)
        assert image_rgb
        assert len(image_rgb._original) == len(image_rgb._images) == len(image_rgb._colors) == n_images

        assert "images=" in repr(image_rgb)

    @staticmethod
    def test_init_fails():
        with pytest.raises(ValueError):
            _ = ImageRGBA(get_images(1)[0])

        with pytest.raises(ValueError):
            _ = ImageRGBA([np.random.randint(0, 255, (10, 10)), np.random.randint(0, 255, (10, 9))])

    @staticmethod
    @pytest.mark.parametrize("color", ("#FF00FF", (1, 0, 1)))
    def test_init_colors_pass(color):
        _ = ImageRGBA(get_images(1), colors=[color])

    @pytest.mark.parametrize("color", ("FFF000", [1, 0], [1, 0, 0, 0, 1]))
    def test_init_colors_fail(self, color):
        with pytest.raises(ValueError):
            _ = ImageRGBA(get_images(1), colors=[color])

    @staticmethod
    @pytest.mark.parametrize("n_images", (1, 3, 5))
    def test_init_images_with_colors(n_images):
        images, colors = get_images(n_images), get_colors(n_images)
        image_rgb = ImageRGBA(images, colors)
        assert image_rgb
        assert len(image_rgb._original) == len(image_rgb._images) == len(image_rgb._colors) == n_images
        for color in colors:
            assert color in image_rgb._colors

    @staticmethod
    def test_init_images_colors_fail():
        with pytest.raises(ValueError):
            images = get_images(1)
            colors = get_colors(2)
            _ = ImageRGBA(images, colors)

    @staticmethod
    @pytest.mark.parametrize("n_images", (1, 3, 5))
    def test_intensities(n_images):
        images = get_images(n_images)
        image_rgb = ImageRGBA(images)

        assert_array_equal(np.sum(images, axis=0), image_rgb.intensities)

    @staticmethod
    @pytest.mark.parametrize("n_images", (1, 3, 5))
    def test_rgb(n_images):
        images = get_images(n_images)
        image_rgb = ImageRGBA(images)

        rgba = image_rgb.rgba
        assert rgba.shape[2] == 4
        assert rgba.max() <= 255
        assert rgba.min() >= 0
        assert rgba.dtype == np.uint8

        rgb = image_rgb.rgb
        assert rgb.shape[2] == 3
        assert rgb.max() <= 255
        assert rgb.min() >= 0
        assert rgb.dtype == np.uint8

        image_rgb.reset()
        assert image_rgb._rgba is None

    @staticmethod
    @pytest.mark.parametrize("n_images", (1, 3, 5))
    @pytest.mark.parametrize("max_value, dtype", ([255, np.uint8], [65535, np.uint16]))
    def test_combine(n_images, max_value, dtype):
        images = get_images(n_images)
        image_rgb = ImageRGBA(images)
        rgba = image_rgb.combine(max_value, dtype)

        assert rgba.shape == (10, 10, 4)
        assert rgba.min() >= 0
        assert rgba.max() <= max_value
        assert rgba.dtype == dtype

    @staticmethod
    @pytest.mark.parametrize("n_images", (1, 3, 5))
    @pytest.mark.parametrize("channel", (0, 1, 2))
    def test_normalize_channel(n_images, channel):
        images = get_images(n_images)
        image_rgb = ImageRGBA(images)

        image_rgb.clip_channel(channel, 20, 100)
        assert image_rgb.rgba[:, :, channel].min() >= 20
        assert image_rgb.rgba[:, :, channel].max() <= 100

    @staticmethod
    @pytest.mark.parametrize("n_images", (1, 3, 5))
    def test_equalize_histogram(n_images):
        images = get_images(n_images)
        image_rgb = ImageRGBA(images)

        image = image_rgb.equalize_histogram()
        assert image.min() >= 0
        assert image.max() <= 255

    @staticmethod
    @pytest.mark.parametrize("n_images", (1, 3, 5))
    @pytest.mark.parametrize("clip_limit", (0.01, 0.5))
    def test_adaptive_histogram(n_images, clip_limit):
        images = get_images(n_images)
        image_rgb = ImageRGBA(images)

        image = image_rgb.adaptive_histogram(clip_limit=clip_limit)
        assert image.min() >= 0
        assert image.max() <= 255

    @staticmethod
    @pytest.mark.parametrize("n_images", (1, 3, 5))
    def test_quantile_rescale(n_images):
        images = get_images(n_images)
        image_rgb = ImageRGBA(images)

        image = image_rgb.quantile_rescale()
        assert image.min() >= 0
        assert image.max() <= 255

    @staticmethod
    @pytest.mark.parametrize("n_images", (1, 3, 5))
    @pytest.mark.parametrize("q_low", (-0.5, 1.01))
    @pytest.mark.parametrize("q_high", (-0.3, 1.5))
    def test_quantile_rescale_fail(n_images, q_low, q_high):
        images = get_images(n_images)
        image_rgb = ImageRGBA(images)

        with pytest.raises(ValueError):
            image_rgb.quantile_rescale(q_low, q_high)

    @staticmethod
    def test_recolor():
        images = get_images(2)
        colors = ["#FF0000", "#00FF00"]
        image_rgb = ImageRGBA(images, colors)
        assert image_rgb._colors[0] == colors[0]
        assert image_rgb._colors[1] == colors[1]
        new_color = "#FFFF00"
        image_rgb.recolor(0, new_color)
        assert image_rgb._colors[0] == new_color

        new_color = [1, 0, 1]
        image_rgb.recolor(0, new_color)
        assert image_rgb._colors[0] == new_color

        with pytest.raises(ValueError):
            image_rgb.recolor(3, new_color)

        with pytest.raises(ValueError):
            image_rgb.recolor(0, (255, 1, 0))

    @staticmethod
    def test_get_one():
        images = get_images(2)
        image_rgb = ImageRGBA(images)

        im = image_rgb.get_one(0, keep_alpha=True)
        assert im.shape[2] == 4
        im = image_rgb.get_one(0, keep_alpha=False)
        assert im.shape[2] == 3

        with pytest.raises(ValueError):
            image_rgb.get_one(3)

    @staticmethod
    @pytest.mark.parametrize("n_bins", (12, 256))
    def test_get_colormap(n_bins):
        images = get_images(2)
        image_rgb = ImageRGBA(images)

        colormap = image_rgb.get_colormap(0, n_bins)
        assert isinstance(colormap, ListedColormap)
        assert len(colormap.colors) == n_bins

        with pytest.raises(ValueError):
            image_rgb.get_colormap(2, n_bins)



"""Image."""
import numpy as np
from bokeh.models import BasicTicker, BoxZoomTool, ColorBar, ColumnDataSource, HoverTool, Range1d
from bokeh.plotting import figure

from plotski.base import Plot
from plotski.utilities import calculate_aspect_ratio


class PlotImageBase(Plot):
    """Basic heatmap plot"""

    # Data attributes
    DATA_KEYS = ("image", "x", "y", "dw", "dh")
    ACTIVE_DRAG = None
    TOOLS = None

    def __init__(
        self,
        output_dir: str,
        source: ColumnDataSource,
        x_axis_label: str = "",
        y_axis_label: str = "",
        title: str = "Heatmap",
        plot_type: str = "heatmap",
        initialize: bool = True,
        **kwargs,
    ):
        self.ACTIVE_DRAG = BoxZoomTool(match_aspect=True)
        self.TOOLS = ("pan, crosshair, reset", self.ACTIVE_DRAG)
        Plot.__init__(
            self,
            output_dir,
            source,
            x_axis_label,
            y_axis_label,
            title=title,
            plot_type=plot_type,
            initilize=initialize,
            **kwargs,
        )

    def plot(self):
        """Main plotting function."""
        raise NotImplementedError("Must implement method")

    def get_figure(self):
        """Get figure."""
        return figure(
            tools=self.kwargs["tools"],
            active_drag=self.kwargs["active_drag"],
            x_range=self.kwargs.get("x_range", Range1d()),
            y_range=self.kwargs.get("y_range", Range1d()),
        )

    def initialize_options(self):
        """Setup few options"""
        from plotski.utilities import convert_colormap_to_mapper

        # setup some common options if the user has not specified them
        if "cmap" not in self.kwargs:
            self.kwargs["cmap"] = "viridis"

        self.kwargs["palette"], self.kwargs["colormapper"] = convert_colormap_to_mapper(
            self.source.data["image"][0],
            self.kwargs["cmap"],
            z_min=self.kwargs.get("z_min", None),
            z_max=self.kwargs.get("z_max", None),
        )
        super().initialize_options()

    def set_hover(self):
        """Set hover."""
        self.figure.add_tools(
            HoverTool(
                show_arrow=True,
                tooltips=[("x, y", "$x{0.00}, $y{0.00}"), (self.kwargs.get("hover_label", "intensity"), "@image")],
            )
        )

    def set_ranges(self, **kwargs):
        """Set ranges."""
        self.figure.xaxis.axis_label_text_baseline = "bottom"

        # update x/y ranges
        src = self.source.data
        if "x_range" not in self.kwargs:
            self.figure.x_range.update(start=0, end=src["image"][0].shape[1])
        if "y_range" not in self.kwargs:
            self.figure.y_range.update(start=0, end=src["image"][0].shape[0])
        # x_range = self.kwargs.get("x_range", None)
        # if x_range is None:
        #     x_range = (0, src["image"][0].shape[1])
        # self.figure.x_range = Range1d(*x_range) if isinstance(x_range, ty.Iterable) else x_range
        # y_range = self.kwargs.get("y_range", None)
        # if y_range is None:
        #     y_range = (0, src["image"][0].shape[0])
        # self.figure.y_range = Range1d(*y_range) if isinstance(y_range, ty.Iterable) else y_range

    def set_figure_dimensions(self):
        """Set figure dimensions."""
        width = self.kwargs.get("width", 600)
        height, width = calculate_aspect_ratio(self.source.data["image"][0].shape, width)
        if height > 600:
            _ratio = 600 / height
            height = 600
            width = int(width * _ratio)
        self.figure.width = self.kwargs.get("width", width)
        self.figure.height = self.kwargs.get("height", height)

    def check_data_source(self):
        """Ensure that each field in the data source is correct"""
        if "image" not in self.source.data:
            raise ValueError("Missing field 'image' in the ColumnDataSource")
        if "image" in self.source.data:
            if not isinstance(self.source.data["image"], list) and len(self.source.data["image"]) > 1:
                raise ValueError("Field 'image' is incorrectly set in ColumnDataSource")
        if "x" not in self.source.data:
            self.source.data["x"] = [0]
        if "y" not in self.source.data:
            self.source.data["y"] = [0]
        if "dw" not in self.source.data:
            self.source.data["dw"] = [self.source.data["image"][0].shape[1]]
        if "dh" not in self.source.data:
            self.source.data["dh"] = [self.source.data["image"][0].shape[0]]
        super().check_data_source()


class PlotImage(PlotImageBase):
    """Image class"""

    def __init__(self, output_dir: str, source: ColumnDataSource, title="Image", **kwargs):
        PlotImageBase.__init__(self, output_dir, source=source, title=title, plot_type="image", **kwargs)

    def plot(self):
        """Plot image."""
        self.plots["image"] = self.figure.image(
            x="x",
            y="y",
            dw="dw",
            dh="dh",
            image="image",
            source=self.source,
            palette=self.kwargs["palette"],
            name="image",
        )
        self.plots["image"].glyph.color_mapper = self.kwargs["colormapper"]

    def add_colorbar(self):
        """Add colorbar."""
        color_bar = ColorBar(
            color_mapper=self.kwargs["colormapper"],
            ticker=BasicTicker(),
            location=(0, 0),
            major_label_text_font_size="10pt",
            label_standoff=8,
        )
        self.figure.add_layout(color_bar, "right")

    def set_options(self):
        """Set options."""
        if self.kwargs.get("add_colorbar", False):
            self.add_colorbar()


class PlotImageRGBA(PlotImageBase):
    """RGB Image class."""

    def __init__(self, output_dir: str, source: ColumnDataSource, title="Image-RGBA", **kwargs):
        PlotImageBase.__init__(self, output_dir, source=source, title=title, plot_type="rgba", **kwargs)

    def plot(self):
        """Main plotting method."""
        self.plots["rgba"] = self.figure.image_rgba(
            x="x", y="y", dw="dw", dh="dh", image="image", source=self.source, name="rgba"
        )

    def check_data_source(self):
        """Check data sources."""
        PlotImageBase.check_data_source(self)
        if self.source.data["image"][0].dtype != np.uint8:
            raise ValueError("ImageRGBA expects 8-bit values")

    def set_hover(self):
        """Add hover to the image plot."""
        tooltips = [("x, y", "$x{0.00}, $y{0.00}")]
        if "intensity" in self.source.data:
            tooltips.append(("intensity", "@intensity"))
        else:
            tooltips.append(("intensity", "@image"))

        if "r" in self.source.data:
            tooltips.append(("(R, G, B A)", "@r, @g, @b, @a"))

        self.figure.add_tools(HoverTool(show_arrow=True, tooltips=tooltips))



"""
Script to replace minimum requirements with exact requirements in setup.cfg

This ensures that our test matrix includes a version with only the minimum
required versions of packages, and we don't accidentally use features only
available in newer versions.

This script does nothing if the 'MIN_REQ' environment variable is anything
other than '1'.
"""
import os
from configparser import ConfigParser
from pathlib import Path


def pin_config_minimum_requirements(config_filename):
    # read it with configparser
    config = ConfigParser()
    config.read(config_filename)

    # swap out >= requirements for ==
    config["options"]["install_requires"] = config["options"]["install_requires"].replace(">=", "==")

    # rewrite setup.cfg with new config
    with open(config_filename, "w") as fout:
        config.write(fout)


if __name__ == "__main__":
    if os.environ.get("MIN_REQ", "") == "1":
        # find setup.cfg
        config_filename = Path(__file__).parent.parent / "setup.cfg"
        if not config_filename.exists():
            print("Config file does not exist.")
        else:
            pin_config_minimum_requirements(str(config_filename))
            print("Pinned minimum requirements in setup.cfg.")



"""Base plot."""
from bokeh.models import ColumnDataSource, HoverTool, Legend, Range1d
from bokeh.plotting import figure

from plotski.base import Plot
from plotski.utilities import check_key


class PlotSpectrum(Plot):
    """Basic Spectrum plot"""

    # Data attributes
    DATA_KEYS = ("x", "y")
    ACTIVE_DRAG = "xbox_zoom"
    TOOLS = ("pan, xpan, xbox_zoom, box_zoom, crosshair, reset",)

    # Defaults
    WIDTH = 800
    HEIGHT = 400

    def __init__(
        self,
        output_dir: str,
        source: ColumnDataSource,
        x_axis_label: str = "x",
        y_axis_label: str = "y",
        title: str = "Spectrum",
        plot_type: str = "spectrum",
        initialize: bool = True,
        **kwargs,
    ):
        Plot.__init__(
            self,
            output_dir,
            source,
            x_axis_label,
            y_axis_label,
            title=title,
            plot_type=plot_type,
            initialize=initialize,
            **kwargs,
        )

    def plot(self):
        """Add plot data"""
        line = self.figure.line(
            x="x",
            y="y",
            source=self.source,
            line_width=self.kwargs["line_width"],
            line_dash=self.kwargs.get("line_dash", "solid"),
            color=self.kwargs["line_color"],
            alpha=self.kwargs["line_alpha"],
            name=self.plot_type,
            legend_label=self.kwargs.get("label", ""),
        )
        self.plots[line.id] = line
        self.add_extents(self.source.data["x"], self.source.data["y"])

    def get_figure(self):
        """Create figure."""
        return figure(
            tools=self.kwargs["tools"],
            active_drag=self.kwargs["active_drag"],
            x_range=self.kwargs.get("x_range", Range1d()),
            y_range=self.kwargs.get("y_range", Range1d()),
        )

    def set_hover(self):
        """Set hover information"""
        self.figure.add_tools(
            HoverTool(
                show_arrow=True,
                tooltips=[(f"{self.metadata['x_axis_label']}", "@x"), (f"{self.metadata['y_axis_label']}", "@y")],
                mode="vline",
                # names=[self.plot_type],
            )
        )

    def initialize_options(self):
        """Convenience function to handle various options set by the user"""
        if "line_width" not in self.kwargs:
            self.kwargs["line_width"] = 1.5
        if "line_color" not in self.kwargs:
            self.kwargs["line_color"] = "#000000"
        if "line_alpha" not in self.kwargs:
            self.kwargs["line_alpha"] = 1.0
        super().initialize_options()

    def set_ranges(self, **kwargs):
        """Set range based on data source"""
        # update x/y ranges
        x_min, x_max, y_min, y_max = self.get_extents(**kwargs)
        if "x_range" not in self.kwargs:
            self.figure.x_range.update(start=x_min, end=x_max)
        if "y_range" not in self.kwargs:
            self.figure.y_range.update(start=y_min, end=y_max)

    def add_plot_line(self, source: ColumnDataSource, **kwargs):
        """Add plot"""
        line = self.figure.line(x="x", y="y", source=source, **kwargs)
        self.plots[line.id] = (source, "Line")
        self.add_extents(source.data["x"], source.data["y"])

    def add_segments(self, source: ColumnDataSource, **kwargs):
        """Add segments"""
        segment = self.figure.segment(x0="x0", y0="y0", x1="x1", y1="y1", source=source, **kwargs)
        self.annotations[segment.id] = (source, "Segment")

    def add_centroids_x(self, source: ColumnDataSource, **kwargs):
        """Add vertical centroids"""
        segment = self.figure.segment(x0="x", y0="y0", x1="x", y1="y1", source=source, **kwargs)
        self.annotations[segment.id] = (source, "Centroid-X")

    def add_centroids_y(self, source: ColumnDataSource, **kwargs):
        """Add horizontal centroids"""
        segment = self.figure.segment(x0="x0", y0="y", x1="x1", y1="y", source=source, **kwargs)
        self.annotations[segment.id] = (source, "Centroid-Y")

    def add_scatter(self, source: ColumnDataSource, **kwargs):
        """Add scatter points."""
        scatter = self.figure.scatter(x="x", y="y", source=source, **kwargs)
        self.annotations[scatter.id] = (source, "Scatter")


class PlotCentroid(PlotSpectrum):
    """Basic centroid plot"""

    DATA_KEYS = ("x", "y0", "y1")

    def __init__(
        self,
        output_dir: str,
        source: ColumnDataSource,
        x_axis_label: str = "x",
        y_axis_label: str = "y",
        title: str = "Centroid Spectrum",
        plot_type: str = "centroid-spectrum",
        **kwargs,
    ):
        PlotSpectrum.__init__(
            self,
            output_dir,
            source=source,
            x_axis_label=x_axis_label,
            y_axis_label=y_axis_label,
            title=title,
            plot_type=plot_type,
            **kwargs,
        )

    def plot(self):
        """Add plot data"""
        label = self.kwargs.get("label", "")
        centroid = self.figure.segment(
            x0="x",
            y0="y0",
            x1="x",
            y1="y1",
            source=self.source,
            line_width=self.kwargs["line_width"],
            color=self.kwargs["line_color"],
            alpha=self.kwargs["line_alpha"],
            name=self.plot_type,
        )
        if label:
            centroid.legend_label = label
        self.plots[centroid.id] = centroid

    def set_hover(self):
        """Set hover"""
        self.figure.add_tools(
            HoverTool(
                show_arrow=True,
                tooltips=[(f"{self.metadata['x_axis_label']}", "@x"), (f"{self.metadata['y_axis_label']}", "@y1")],
                point_policy="snap_to_data",
                line_policy="none",
            )
        )

    def set_ranges(self, **kwargs):
        """Set ranges"""
        # update x/y ranges
        src = self.source.data
        if "x_range" not in self.kwargs:
            self.figure.x_range = Range1d(min(src["x"]), max(src["x"]))
        if "y_range" not in self.kwargs:
            self.figure.y_range = Range1d(min(src["y0"]), max(src["y1"]) * 1.05)


class PlotButterflySpectrum(PlotSpectrum):
    """Butterfly plot"""

    DATA_KEYS = ("x_top", "y_top", "x_bottom", "y_bottom")

    def __init__(
        self,
        output_dir: str,
        source: ColumnDataSource,
        x_axis_label: str = "x",
        y_axis_label: str = "y",
        title: str = "Butterfly Spectrum",
        plot_type: str = "butterfly-spectrum",
        **kwargs,
    ):
        PlotSpectrum.__init__(
            self,
            output_dir,
            source=source,
            x_axis_label=x_axis_label,
            y_axis_label=y_axis_label,
            title=title,
            plot_type=plot_type,
            **kwargs,
        )

    def plot(self):
        """Plot data"""
        line_top = self.figure.line(
            x="x_top",
            y="y_top",
            source=self.source,
            line_width=self.kwargs["line_width"],
            color=self.kwargs["line_color"],
            alpha=self.kwargs["line_alpha"],
            name=self.plot_type + "-top",
        )
        self.plots[line_top.id] = line_top
        line_bottom = self.figure.line(
            x="x_bottom",
            y="y_bottom",
            source=self.source,
            line_width=self.kwargs["line_width"],
            color=self.kwargs["line_color"],
            alpha=self.kwargs["line_alpha"],
            name=self.plot_type + "-bottom",
        )
        self.plots[line_bottom.id] = line_bottom

    def add_legend(self):
        """Add legend item to the plot"""
        legend = Legend(
            items=[
                ("Top", self.figure.select(name=self.plot_type + "-top")),
                ("Bottom", self.figure.select(name=self.plot_type + "-bottom")),
            ],
            orientation="horizontal",
        )
        # Add the layout outside the plot, clicking legend item hides the line
        self.figure.add_layout(legend, "above")
        self.figure.legend.click_policy = "hide"

    def set_ranges(self, **kwargs):
        """Set ranges"""
        # update x/y ranges
        src = self.source.data
        x = [min(src["x_top"]), min(src["x_bottom"]), max(src["x_top"]), max(src["x_bottom"])]
        y = [min(src["y_top"]), min(src["y_bottom"]), max(src["y_top"]), max(src["y_bottom"])]
        self.figure.x_range = Range1d(min(x), max(x))
        self.figure.y_range = Range1d(min(y) * 1.05, max(y) * 1.05)

    def set_hover(self):
        """Set hover"""
        self.figure.add_tools(
            HoverTool(
                show_arrow=True,
                tooltips=[
                    (f"{self.x_axis_label}", "@x_top"),
                    (f"{self.y_axis_label}", "@y_top"),
                ],
                renderers=self.figure.select(name=self.plot_type + "-top"),
                mode="vline",
            )
        )
        self.figure.add_tools(
            HoverTool(
                show_arrow=True,
                tooltips=[
                    (f"{self.x_axis_label}", "@x_bottom"),
                    (f"{self.y_axis_label}", "@y_bottom"),
                ],
                renderers=self.figure.select(name=self.plot_type + "-bottom"),
                mode="vline",
            )
        )


class PlotMultiLine(PlotSpectrum):
    """Basic multiline spectrum"""

    DATA_KEYS = ("xs", "ys")

    def __init__(
        self,
        output_dir: str,
        source: ColumnDataSource,
        x_axis_label: str = "x",
        y_axis_label: str = "y",
        title: str = "Multi-line",
        plot_type: str = "multiline-spectrum",
        **kwargs,
    ):
        PlotSpectrum.__init__(
            self,
            output_dir,
            source=source,
            x_axis_label=x_axis_label,
            y_axis_label=y_axis_label,
            title=title,
            plot_type=plot_type,
            **kwargs,
        )

    def plot(self):
        """Plot data"""
        multiline = self.figure.multi_line(
            xs="xs",
            ys="ys",
            source=self.source,
            line_width=self.kwargs["line_width"],
            color=self.kwargs["line_color"] if not check_key(self.source, "colors") else "colors",
            alpha=self.kwargs["line_alpha"] if not check_key(self.source, "alpha") else "alpha",
            name=self.plot_type,
        )
        self.plots[multiline.id] = multiline

    def set_ranges(self, **kwargs):
        """Set plot ranges"""
        src = self.source.data
        xmin = min([min(x) for x in src["xs"]])
        xmax = max([max(x) for x in src["xs"]])
        ymin = min([min(y) for y in src["ys"]])
        ymax = max([max(y) for y in src["ys"]])
        self.figure.x_range.update(start=xmin, end=xmax)
        self.figure.y_range.update(start=ymin, end=ymax)
        # x = [min(src["x_top"]), min(src["x_bottom"]), max(src["x_top"]), max(src["x_bottom"])]
        # y = [min(src["y_top"]), min(src["y_bottom"]), max(src["y_top"]), max(src["y_bottom"])]
        # self.figure.x_range = Range1d(min(x), max(x))
        # self.figure.y_range = Range1d(min(y) * 1.05, max(y) * 1.05)

    def set_hover(self):
        """Set hover"""
        tooltips = [
            (f"{self.x_axis_label}", "x"),
            (f"{self.y_axis_label}", "y"),
        ]
        if check_key(self.source, "line_id"):
            tooltips.append(("Line ID", "@line_id"))

        self.figure.add_tools(
            HoverTool(
                show_arrow=True,
                tooltips=tooltips,
                line_policy="next",
            )
        )



"""Custom plots."""
from bokeh.models import ColumnDataSource

from plotski.spectrum.plot import PlotButterflySpectrum, PlotCentroid, PlotSpectrum


class PlotCentroidMassSpectrum(PlotCentroid):
    """Plot centroid mass spectrum"""

    def __init__(
        self,
        output_dir: str,
        source: ColumnDataSource,
        x_axis_label="m/z",
        y_axis_label="Intensity",
        title="Centroid Mass Spectrum",
        **kwargs,
    ):
        PlotCentroid.__init__(
            self,
            output_dir,
            source=source,
            x_axis_label=x_axis_label,
            y_axis_label=y_axis_label,
            title=title,
            plot_type="centroid-mass-spectrum",
            **kwargs,
        )


class PlotMassSpectrum(PlotSpectrum):
    """Mass spectrum plot"""

    def __init__(
        self,
        output_dir: str,
        source: ColumnDataSource,
        x_axis_label: str = "m/z",
        y_axis_label: str = "Intensity",
        title: str = "Mass Spectrum",
        **kwargs,
    ):
        PlotSpectrum.__init__(
            self,
            output_dir,
            source=source,
            x_axis_label=x_axis_label,
            y_axis_label=y_axis_label,
            title=title,
            plot_type="mass-spectrum",
            **kwargs,
        )


class PlotMobilogram(PlotSpectrum):
    """Mobilogram plot"""

    def __init__(
        self,
        output_dir: str,
        source: ColumnDataSource,
        x_axis_label: str = "Drift time (bins)",
        y_axis_label: str = "Intensity",
        title: str = "Mobilogram",
        **kwargs,
    ):
        PlotSpectrum.__init__(
            self,
            output_dir,
            source,
            x_axis_label=x_axis_label,
            y_axis_label=y_axis_label,
            title=title,
            plot_type="mobilogram",
            **kwargs,
        )


class PlotButterflyMassSpectrum(PlotButterflySpectrum):
    """Make butterfly mass spectrum"""

    def __init__(
        self,
        output_dir: str,
        source: ColumnDataSource,
        x_axis_label: str = "m/z",
        y_axis_label: str = "Intensity",
        title: str = "Butterfly Mass Spectrum",
        **kwargs,
    ):
        PlotButterflySpectrum.__init__(
            self,
            output_dir,
            source=source,
            x_axis_label=x_axis_label,
            y_axis_label=y_axis_label,
            title=title,
            plot_type="butterfly-mass-spectrum",
            **kwargs,
        )


class PlotButterflyMobilogram(PlotButterflySpectrum):
    """Make butterfly mobilogram"""

    def __init__(
        self,
        output_dir: str,
        source: ColumnDataSource,
        x_axis_label: str = "Drift time (bins)",
        y_axis_label: str = "Intensity",
        title: str = "Butterfly Mobilogram",
        **kwargs,
    ):
        PlotButterflySpectrum.__init__(
            self,
            output_dir,
            source=source,
            x_axis_label=x_axis_label,
            y_axis_label=y_axis_label,
            title=title,
            plot_type="butterfly-mobilogram",
            **kwargs,
        )



"""Bokeh plot store"""
import os
import typing as ty
import warnings
import webbrowser

import numpy as np
from bokeh.io import save
from bokeh.layouts import column, gridplot, row
from bokeh.models import ColumnDataSource

try:
    from bokeh.models import TabPanel
except ImportError:
    from bokeh.models.widgets import Panel as TabPanel

try:
    from bokeh.models import Tabs
except ImportError:
    from bokeh.models.widgets import Tabs

from plotski.base import Plot
from plotski.image import PlotImage, PlotImageRGBA
from plotski.scatter import PlotScatter
from plotski.spectrum.plot import PlotCentroid, PlotMultiLine, PlotSpectrum
from plotski.store.containers import Column, Container, Grid, Individual, Row
from plotski.utilities import get_unique_str

# TODO: add repr that shows the layout of the store e.g. tab 1 \ plot 1 plot 2 plot 3; tab 2 \ plot 1 plot 2 plot 3
# TODO: add option to annotate spectrum and heatmap with rois and/or peaks


class PlotStore:
    """Main class that is responsible for managing all plots that should be exported to static HTML document."""

    def __init__(self, output_dir: str = "", options=None, filename="figure-store.html", title: str = "Document Store"):
        self.output_dir = output_dir
        self.filename = filename
        self.options = options

        # setup document parameters
        self.document_title = title

        # store of figures
        self.tabs: ty.Dict[str, ty.Dict[str, Container[Plot]]] = {}

    def __repr__(self) -> str:
        """Print"""
        return f"PlotStore <tabs={len(self.tabs)}>"

    def __getitem__(self, tab: ty.Union[int, str]):
        """Get tab object"""
        if isinstance(tab, int):
            tab = self.tab_names[tab]
        if tab not in self.tabs:
            raise KeyError(f"Tab '{tab}' does not exist")
        return self.tabs[tab]

    def __len__(self) -> int:
        """Get number of tabs"""
        return len(self.tabs)

    @property
    def tab_names(self) -> ty.List[str]:
        """Get list of tab names"""
        return list(self.tabs.keys())

    def check_tab(self, tab_name: str, auto_add: bool = True):
        """Check whether tab exists in the container"""
        if tab_name not in self.tabs:
            if auto_add:
                self.add_tab(tab_name)
            else:
                raise ValueError(f"Missing tab {tab_name}")

    @staticmethod
    def check_data(data: ty.Dict, keys: ty.Iterable):
        """Check whether data is a dictionary and contains appropriate data keys."""
        if not isinstance(data, dict):
            raise ValueError("Provided value of `data` is incorrect. Expected dictionary.")
        missing = []
        for key in keys:
            if key not in data:
                missing.append(key)
        if missing:
            raise ValueError(
                f"Provided data is missing some keys. Expected: {', '.join(keys)}."
                f" Missing keys: {', '.join(missing)}"
            )

    def show(self, tab_names=None, always_as_tabs: bool = True):
        """Return HTML representation of the document"""
        from bokeh.io import show

        return show(self.get_layout(tab_names, always_as_tabs))

    def get_layout(self, tab_names: ty.Optional[ty.List[str]] = None, always_as_tabs: bool = True):
        """Return fully ordered Bokeh document which can be visualised (using 'show' command) or exported as HTML

        Parameters
        ----------
        tab_names : list
            list of tab names which must be present in the `tabs` container
        always_as_tabs : bool
            if 'True', the resultant HTML document will contain 'Tabs' even if only one tab is present

        Returns
        -------
        tabs : Tabs
            Tabs container

        """

        def unpack_figures():
            """Unpack layout elements from item/row/column or grid"""
            return [plot.layout for plot in item_contents]

        # user can specify which tabs they would like to export as HTML document. If 'tab_names' was not specified,
        # we will use all tabs in the exported document
        if tab_names is None:
            tab_names = self.tab_names
        else:
            # let's check if the user supplied single tab name using strings...
            if isinstance(tab_names, str):
                tab_names = [tab_names]

        # check that each tab name is actually present in the tab store
        if not all([tab_name in self.tabs for tab_name in tab_names]):
            raise ValueError("Some of the specified tab names are not present in the figure store")

        # initialize panel store
        panels = []
        # iterate over each tab and append the contents. Here, each panel corresponds to single tab
        for tab_name in tab_names:
            tab_contents = self.tabs[tab_name]
            _tab_contents = []
            # iterate over each object specified in the tab
            for item_name, item_contents in tab_contents.items():
                # items can be specified as an 'item' (single element)
                figures = unpack_figures()
                if isinstance(item_contents, Individual):
                    _tab_contents.extend(figures)
                # row (multiple elements in a row)
                elif isinstance(item_contents, Row):
                    _tab_contents.append(row(figures))
                # column (multiple elements in a column)
                elif isinstance(item_contents, Column):
                    _tab_contents.append(column(figures))
                # grid (multiple elements in a grid):
                elif isinstance(item_contents, Grid):
                    _tab_contents.append(gridplot(figures, ncols=item_contents.n_cols))

            if not _tab_contents:
                print("Tab was empty - not adding it into the HTML document")
                continue
            panels.append(TabPanel(child=column(children=_tab_contents), title=tab_name))

        # if the 'always_as_tabs' toggle is disabled and only one tab is present, the returned object will be column
        # element
        if len(panels) == 1 and not always_as_tabs:
            return panels[0].child

        return Tabs(tabs=panels)

    def save(self, filepath=None, show=True, **kwargs) -> str:
        """Save Bokeh document as HTML file

        Parameters
        ----------
        filepath : Path, optional
            path where to save the HTML document
        show : bool
            if 'True', newly generated document will be shown in the browser
        kwargs :
            parameters to be passed on to the 'get_layout' function
        """

        if filepath is None:
            filepath = os.path.join(self.output_dir, self.filename)

        with warnings.catch_warnings():
            warnings.filterwarnings("ignore")
            save(self.get_layout(**kwargs), filepath, title=self.document_title)

        # open figure in browser
        if show:
            webbrowser.open_new_tab(filepath)
        return filepath

    def get_unique_name(self, tab_name: str, basename: str = "item"):
        """Get unique name for an item in specific tab. Names are made unique by adding #NUMBER+1 itself

        Parameters
        ----------
        tab_name : str
            name of the tab
        basename : str
            base name of the unique name, default = 'item'

        Returns
        -------

        """
        i = 0
        while f"{basename} #{i}" in self.tabs[tab_name]:
            i += 1
        return f"{basename} #{i}"

    def add_tab(self, tab_name: str, override=False) -> str:
        """Add new tab to the document

        Parameters
        ----------
        tab_name : str
            name of the tab to be added to the document
        override : bool
            if tab with name `tab_name` already exists and `reset` is set to `True` no exception will be thrown

        Returns
        -------
        tab_name : str
            name of the new tab

        Raises
        ------
        ValueError
            raised if `tab_name` already present in the `self.tabs` container. It will not be thrown if `reset` is set
            to `True
        """
        if tab_name in self.tabs and not override:
            raise ValueError(
                "This tab has previously been added! Please set 'reset' to True if you would like to"
                " override current container."
            )
        self.tabs[tab_name] = {}
        return tab_name

    def add_tabs(self, tab_names: ty.List[str], override=False):
        """Add multiple new tabs to the document

        Parameters
        ----------
        tab_names : List[str]
            List of tab names.
        override : bool, optional
            if tab with name `tab_name` already exists and `reset` is set to `True` no exception will be thrown

        Raises
        ------
        ValueError
            Raised if `tab_name` already present in the `self.tabs` container. It will not be thrown if `reset` is set
            to `True
        """
        for tab_name in tab_names:
            self.add_tab(tab_name, override)

    def add_row(self, tab_name: str) -> str:
        """Add row to particular tab.

        Parameters
        ----------
        tab_name : str
            Name of the tab where row should be added.

        Returns
        -------
        row_name : str
            Name of the generated row container. Name is automatically generated and always unique.
        """
        assert tab_name in self.tabs

        row_name = self.get_unique_name(tab_name, "row")
        self.tabs[tab_name][row_name] = Row()
        return row_name

    def add_col(self, tab_name: str) -> str:
        """Add column to particular tab.

        Parameters
        ----------
        tab_name : str
            Name of the tab where row should be added.

        Returns
        -------
        col_name : str
            Name of the generated column container. Name is automatically generated and always unique.
        """
        assert tab_name in self.tabs

        col_name = self.get_unique_name(tab_name, "col")
        self.tabs[tab_name][col_name] = Column()
        return col_name

    def add_grid(self, tab_name: str, n_cols: ty.Optional[int] = None) -> str:
        """Add grid to particular tab.

        Parameters
        ----------
        tab_name : str
            Name of the tab where grid should be added.
        n_cols : int, optional
            Number of columns the grid should use during rendering. If value is not provided, value will be calculated
            by rounding square root of the number of contained figures.

        Returns
        -------
        grid_name : str
            Name of the generated grid container. Name is automatically generated and always unique.
        """
        assert tab_name in self.tabs

        grid_name = self.get_unique_name(tab_name, "grid")
        grid = Grid()
        grid.n_cols = n_cols
        self.tabs[tab_name][grid_name] = grid
        return grid_name

    def append_item(self, tab_name: str, layout_name: str, plot: Plot):
        """Append plot object to tab/item_name so it can be easily retrieved later on

        Parameters
        ----------
        tab_name : str
            Name of the tab where plot should be added to.
        layout_name : str
            Name of the layout where plot item should be added to. The layout item is present in particular tab.
        plot : Plot
            Plot object that should be added to the layout.
        """
        assert tab_name in self.tabs, f"Tab {tab_name} not found in tabs. Make sure to add it first."
        if layout_name not in self.tabs[tab_name]:
            self.tabs[tab_name][layout_name] = Individual()

        # set the plot name
        plot.name = get_unique_str()
        self.tabs[tab_name][layout_name].append(plot)
        return plot

    def plot_scatter(self, tab_name, data: ty.Dict, layout_name=None, **kwargs):
        """Adds generic scatter to the plot store

        Parameters
        ----------
        tab_name : str
            Name of the tab where plot should be added to.
        data : dict
            Dictionary containing appropriate plot fields, in this case:
                x, y = list / array
            the length of x and y must be the same
        layout_name : str
            By default, plot objects are added to the tab in iterative way (e.g. if there are no plots in the tab, it
            will be added as 'item #0', if there is one then it will be added as 'item #1' etc. Sometimes you might want
            to add it to a 'row' or 'column' for which you have name - you can specify its name here and if its present
            the plot object will be added to that container.
        kwargs :
            dictionary containing plot parameters e.g. x/y axis labels, title, etc...

        Returns
        -------
        tab_name : str
            name of the tab
        item_name : str
            name of the plot
        plot : Plot
            plot object
        """
        self.check_tab(tab_name)
        self.check_data(data, ("x", "y"))

        source = ColumnDataSource(data)
        plot = PlotScatter(self.output_dir, source=source, **kwargs)

        # add figure object to tab
        layout_name = layout_name if layout_name is not None else self.get_unique_name(tab_name)
        self.append_item(tab_name, layout_name, plot)
        return tab_name, layout_name, plot

    def plot_spectrum(self, tab_name, data: ty.Dict, layout_name=None, **kwargs):
        """Adds generic spectrum to the plot store

        Parameters
        ----------
        tab_name : str
            name of the tab where plot should be added to
        data : dict
            Dictionary containing appropriate plot fields, in this case:
                x, y = list / array
            the length of x and y must be the same
        layout_name : str
            By default, plot objects are added to the tab in iterative way (e.g. if there are no plots in the tab, it
            will be added as 'item #0', if there is one then it will be added as 'item #1' etc. Sometimes you might want
            to add it to a 'row' or 'column' for which you have name - you can specify its name here and if its present
            the plot object will be added to that container.
        kwargs :
            dictionary containing plot parameters e.g. x/y axis labels, title, etc...

        Returns
        -------
        tab_name : str
            name of the tab
        item_name : str
            name of the plot
        plot : PlotSpectrum
            plot object
        """
        self.check_tab(tab_name)
        self.check_data(data, ("x", "y"))

        source = ColumnDataSource(data)
        plot = PlotSpectrum(self.output_dir, source=source, **kwargs)

        # add figure object to tab
        layout_name = layout_name if layout_name is not None else self.get_unique_name(tab_name)
        self.append_item(tab_name, layout_name, plot)
        return tab_name, layout_name, plot

    def plot_centroids_x(self, tab_name, data: ty.Dict, layout_name=None, **kwargs):
        """Adds generic spectrum to the plot store

        Parameters
        ----------
        tab_name : str
            name of the tab where plot should be added to
        data : dict
            Dictionary containing appropriate plot fields, in this case:
                x, y0, y1 = list / array
            the length of x and y0 and y1 must be the same.
            If y0 is not provided, values will be automatically generated to provide array of 0s
        layout_name : str
            By default, plot objects are added to the tab in iterative way (e.g. if there are no plots in the tab, it
            will be added as 'item #0', if there is one then it will be added as 'item #1' etc. Sometimes you might want
            to add it to a 'row' or 'column' for which you have name - you can specify its name here and if its present
            the plot object will be added to that container.
        kwargs :
            dictionary containing plot parameters e.g. x/y axis labels, title, etc...

        Returns
        -------
        tab_name : str
            name of the tab
        item_name : str
            name of the plot
        plot : PlotCentroid
            plot object
        """
        self.check_tab(tab_name)
        if "y0" not in data:
            data["y0"] = np.zeros_like(data["x"], dtype=np.int8)
        self.check_data(data, ("x", "y0", "y1"))

        source = ColumnDataSource(data)
        plot = PlotCentroid(self.output_dir, source=source, **kwargs)

        # add figure object to tab
        layout_name = layout_name if layout_name is not None else self.get_unique_name(tab_name)
        self.append_item(tab_name, layout_name, plot)
        return tab_name, layout_name, plot

    def plot_centroids_y(self, tab_name, data: ty.Dict, layout_name=None, **kwargs):
        """Adds generic spectrum to the plot store

        Parameters
        ----------
        tab_name : str
            name of the tab where plot should be added to
        data : dict
            Dictionary containing appropriate plot fields, in this case:
                x0, x1, y = list / array
            the length of x and y0 and y1 must be the same.
            If x0 is not provided, values will be automatically generated to provide array of 0s
        layout_name : str
            By default, plot objects are added to the tab in iterative way (e.g. if there are no plots in the tab, it
            will be added as 'item #0', if there is one then it will be added as 'item #1' etc. Sometimes you might want
            to add it to a 'row' or 'column' for which you have name - you can specify its name here and if its present
            the plot object will be added to that container.
        kwargs :
            dictionary containing plot parameters e.g. x/y axis labels, title, etc...

        Returns
        -------
        tab_name : str
            name of the tab
        item_name : str
            name of the plot
        plot : PlotCentroid
            plot object
        """
        self.check_tab(tab_name)
        if "x0" not in data:
            data["x0"] = np.zeros_like(data["y"], dtype=np.int8)
        self.check_data(data, ("x0", "x1", "y"))

        source = ColumnDataSource(data)
        plot = PlotCentroid(self.output_dir, source=source, **kwargs)

        # add figure object to tab
        layout_name = layout_name if layout_name is not None else self.get_unique_name(tab_name)
        self.append_item(tab_name, layout_name, plot)
        return tab_name, layout_name, plot

    def plot_multiline_spectrum(self, tab_name, data: ty.Dict, layout_name=None, **kwargs):
        """Adds multiple-lines to the same plot area

        Parameters
        ----------
        tab_name : str
            name of the tab where plot should be added to
        data : dict
            Dictionary containing appropriate plot fields, in this case:
                xs, ys = list / array
            the length of x and y must be the same
        layout_name : str
            By default, plot objects are added to the tab in iterative way (e.g. if there are no plots in the tab, it
            will be added as 'item #0', if there is one then it will be added as 'item #1' etc. Sometimes you might want
            to add it to a 'row' or 'column' for which you have name - you can specify its name here and if its present
            the plot object will be added to that container.
        kwargs :
            dictionary containing plot parameters e.g. x/y axis labels, title, etc...

        Returns
        -------
        tab_name : str
            name of the tab
        item_name : str
            name of the plot
        plot : PlotMultiLine
            plot object
        """
        self.check_tab(tab_name)
        self.check_data(data, ("xs", "ys"))

        source = ColumnDataSource(data)
        plot = PlotMultiLine(self.output_dir, source=source, **kwargs)

        # add figure object to tab
        layout_name = layout_name if layout_name is not None else self.get_unique_name(tab_name)
        self.append_item(tab_name, layout_name, plot)
        return tab_name, layout_name, plot

    def plot_image(self, tab_name, data: ty.Dict, layout_name=None, **kwargs):
        """Adds image to the plot store

        Parameters
        ----------
        tab_name : str
            name of the tab where plot should be added to
        data : dict
            Dictionary containing appropriate plot fields, in this case:
                image = list with single array
            If the `image` item is not provided as list with single array then it will be automatically placed there.
        layout_name : str
            by default, plot objects are added to the tab in iterative way (e.g. if there are no plots in the tab, it
            will be added as 'item #0', if there is one then it will be added as 'item #1' etc. Sometimes you might want
            to add it to a 'row' or 'column' for which you have name - you can specify its name here and if its present
            the plot object will be added to that container
        kwargs :
            dictionary containing plot parameters e.g. x/y axis labels, title, etc...

        Returns
        -------
        tab_name : str
            name of the tab
        item_name : str
            name of the plot
        plot : PlotImage
            plot object
        """
        self.check_tab(tab_name)
        self.check_data(data, ("image",))
        # Bokeh expects nested list to plat an image
        if isinstance(data["image"], np.ndarray):
            data["image"] = [data["image"]]

        source = ColumnDataSource(data)
        plot = PlotImage(self.output_dir, source=source, **kwargs)

        # add figure object to tab
        layout_name = layout_name if layout_name is not None else self.get_unique_name(tab_name)
        self.append_item(tab_name, layout_name, plot)
        return tab_name, layout_name, plot

    def plot_rgb_image(self, tab_name, data: ty.Dict, layout_name=None, **kwargs):
        """Adds RGBA image to the plot store

        Parameters
        ----------
        tab_name : str
            name of the tab where plot should be added to
        data : dict
            dictionary containing appropriate plot fields
            in this case:
                image = 3D array
            the 'image' item must be embedded in a list otherwise you will be greeted with nasty exception
        layout_name : str
            by default, plot objects are added to the tab in iterative way (e.g. if there are no plots in the tab, it
            will be added as 'item #0', if there is one then it will be added as 'item #1' etc. Sometimes you might want
            to add it to a 'row' or 'column' for which you have name - you can specify its name here and if its present
            the plot object will be added to that container
        kwargs :
            dictionary containing plot parameters e.g. x/y axis labels, title, etc...

        Returns
        -------
        tab_name : str
            name of the tab
        item_name : str
            name of the plot
        plot : Plot
            plot object
        """
        self.check_tab(tab_name)
        self.check_data(data, ("image",))
        # Bokeh expects nested list to plat an image
        if isinstance(data["image"], np.ndarray):
            data["image"] = [data["image"]]

        source = ColumnDataSource(data)
        plot = PlotImageRGBA(self.output_dir, source=source, **kwargs)

        # add figure object to tab
        layout_name = layout_name if layout_name is not None else self.get_unique_name(tab_name)
        self.append_item(tab_name, layout_name, plot)
        return tab_name, layout_name, plot

    def add_line_plot(self, plot, data: ty.Dict, **kwargs):
        """Adds generic spectrum to the plot store

        Parameters
        ----------
        plot : PlotSpectrum
            plot object where line spectrum should be added to
        data : dict
            Dictionary containing appropriate plot fields, in this case:
                x, y = list / array
            the length of x and y must be the same
        kwargs :
            dictionary containing plot parameters e.g. x/y axis labels, title, etc...

        Returns
        -------
        tab_name : str
            name of the tab
        item_name : str
            name of the plot
        plot : Plot
            plot object
        """
        self.check_data(data, ("x", "y"))
        source = ColumnDataSource(data)
        plot.add_plot_line(source, **kwargs)

    def add_band(self, plot, data: ty.Dict, **kwargs):
        """Add band to the plot area to highlight specific region, display standard deviation of display errors

        Parameters
        ----------
        plot : Plot
            plot object to add the band to
        data : dict
            dictionary containing appropriate plot fields
            in this case:
                base = the orthogonal coordinates of the upper and lower values
                lower = the coordinates of the lower portion of the filled area band
                upper = the coordinates of the upper portion of the filled area band
        kwargs :
            dictionary containing plot parameters e.g. line width, line color, transparency, etc...
            must be valid Bokeh fields

        Examples
        --------
        >>> import numpy as np
        >>> x = np.arange(10)
        >>> y = np.arange(10)
        >>> store = PlotStore("")
        >>> _, _, plot = store.plot_spectrum("plot", dict(x=x, y=y))
        >>> store.add_band(plot, dict(base=x, lower=y-3, upper=y+3))
        """
        if not hasattr(plot, "add_band"):
            raise ValueError("Cannot add band to this plot")
        self.check_data(data, ("base", "lower", "upper"))
        source = ColumnDataSource(data)
        plot.add_band(source, **kwargs)

    def add_span(self, plot, data: ty.Dict, **kwargs):
        """Add span line(s) to the plot area

        You can specify multiple lines simultaneously by setting the `location` key to an iterable. In case multiple
        values are provided, they will have the same `dimension`

        Parameters
        ----------
        plot : Plot
            plot object to add the band to
        data : dict
            dictionary containing appropriate plot fields
            in this case:
                location = the location of the span along `dimension`
                dimension = the direction of the span. Can be either `height` (y-direction) or `width` (x-direction)
        kwargs :
            dictionary containing plot parameters e.g. line width, line color, transparency, etc...
            must be valid Bokeh fields

        Examples
        --------
        >>> import numpy as np
        >>> x = np.arange(10)
        >>> y = np.arange(10)
        >>> store = PlotStore("")
        >>> _, _, plot = store.plot_spectrum("plot", dict(x=x, y=y))
        >>> store.add_span(plot, dict(location=1, dimension="width"))
        >>> store.add_span(plot, dict(location=1, dimension="height"))
        >>> store.add_span(plot, dict(location=[5, 3], dimension="height"))
        """
        if not hasattr(plot, "add_span"):
            raise ValueError("Cannot add band to this plot")
        self.check_data(data, ("location", "dimension"))
        plot.add_span(data, **kwargs)

    def add_box(self, plot, data: ty.Dict, **kwargs):
        """Add box to the plot area to highlight region of interest

        Parameters
        ----------
        plot : Plot
            plot object to add the band to
        data : dict
            dictionary containing appropriate plot fields - not all of the these need to be defined, so for instance
            specifying `bottom` will result in a box annotation that spans from the `bottom` to the edge of the plot
            area and if you specify `top` and `left` will result in box annotation that starts at `top` and `left`
            and spans to the edge of the plot. Specifying all four attributes will result box of the size that
            highlights specifying area and might not span the entire plot area
            in this case:
                bottom = the y-coordinates of the bottom edge of the box annotation
                top = the y-coordinates of the top edge of  the box annotation
                left = the x-coordinates of the left edge of the box annotation
                right = the x-coordinates of the right edge of the box annotation
        kwargs :
            dictionary containing plot parameters e.g. line width, line color, transparency, etc...
            must be valid Bokeh fields

        Examples
        --------
        >>> import numpy as np
        >>> x = np.arange(10)
        >>> y = np.arange(10)
        >>> store = PlotStore("")
        >>> _, _, plot = store.plot_spectrum("plot", dict(x=x, y=y))
        >>> store.add_box(plot, dict(bottom=1, right=3, top=4, left=2), line_width=5, line_color="red")
        >>> store.add_box(plot, dict(bottom=3, right=7))
        >>> store.add_box(plot, dict(top=4, left=2))
        """
        if not hasattr(plot, "add_box"):
            raise ValueError("Cannot add box to this plot")
        self.check_data(data, ("bottom", "top", "left", "right"))
        plot.add_box(data, **kwargs)

    def add_patch(self, plot, data: ty.List[ty.List], **kwargs):
        """Add patch/polygon to the plot area to highlight region of interest

        Parameters
        ----------
        plot : Plot
            plot object to add the band to
        data : list of lists
            formatted list of x and y coordinates in format [[x1, x2, x3], [y1, y2, y3]]
        kwargs :
            dictionary containing plot parameters e.g. line width, line color, transparency, etc...
            must be valid Bokeh fields
        """
        assert isinstance(data, list)
        if not hasattr(plot, "add_patch"):
            raise ValueError("Cannot add box to this plot")
        plot.add_patch(data, **kwargs)

    def add_labels(self, plot, data: ty.Dict, **kwargs):
        """Add label set to an plot/image

        Parameters
        ----------
        plot : Plot
            plot object to add the band to
        data : dict
            dictionary containing appropriate plot fields
            in this case:
                x = the x-coordinates of the labels
                y = the y-coordinates of the labels
                text = the labels that go with the x/y coordinates
        kwargs :
            dictionary containing plot parameters e.g. line width, line color, transparency, etc...
            must be valid Bokeh fields

        Examples
        --------
        >>> import numpy as np
        >>> x = np.arange(10)
        >>> y = np.arange(10)
        >>> store = PlotStore("")
        >>> _, _, plot = store.plot_spectrum("plot", dict(x=x, y=y))
        >>> store.add_labels(plot, dict(x=[3, 4], y=[3, 4], text=["label 1", "label 2"]))
        """
        assert isinstance(data, dict)
        if not hasattr(plot, "add_labels"):
            raise ValueError("Cannot add band to this plot")
        self.check_data(data, ("x", "y", "text"))
        source = ColumnDataSource(data)
        plot.add_labels(source, **kwargs)

    def add_segments(self, plot: PlotSpectrum, data: ty.Dict, **kwargs):
        """Add label set to an plot/image

        Parameters
        ----------
        plot : PlotSpectrum
            plot object to add the band to
        data : dict
            dictionary containing appropriate plot fields
            in this case:
                x0 = the start of the x-coordinates for each line
                x1 = the end of the x-coordinates for each line
                y0 = the start of the y-coordinates for each vertical line
                y1 = the end of the y-coordinates for each vertical line
        kwargs :
            dictionary containing plot parameters e.g. line width, line color, transparency, etc...
            must be valid Bokeh fields

        Examples
        --------
        >>> import numpy as np
        >>> x = np.arange(10)
        >>> y = np.arange(10)
        >>> store = PlotStore("")
        >>> _, _, plot = store.plot_spectrum("plot", dict(x=x, y=y))
        >>> store.add_segments(plot, dict(x0=[0, 9], x1=[9, 0], y0=[0, 0], y1=[10, 10]))
        """
        if not hasattr(plot, "add_segments"):
            raise ValueError("Cannot add segments to this plot")
        self.check_data(data, ("x0", "x1", "y0", "y1"))
        source = ColumnDataSource(data)
        plot.add_segments(source, **kwargs)

    def add_centroids_x(self, plot: PlotSpectrum, data: ty.Dict, **kwargs):
        """Add vertical centroids/lines to a particular plot

        Parameters
        ----------
        plot : PlotSpectrum
            plot object to add the band to
        data : dict
            dictionary containing appropriate plot fields
            in this case:
                x = the x-coordinates for each for each line
                y0 = the start of the y-coordinates for each vertical line
                y1 = the end of the y-coordinates for each vertical line
        kwargs :
            dictionary containing plot parameters e.g. line width, line color, transparency, etc...
            must be valid Bokeh fields

        Examples
        --------
        >>> import numpy as np
        >>> x = np.arange(10)
        >>> y = np.arange(10)
        >>> store = PlotStore("")
        >>> _, _, plot = store.plot_spectrum("plot", dict(x=x, y=y))
        >>> store.add_centroids_x(plot, dict(x=[1, 2, 3], y0=[0, 0, 0], y1=[3, 5, 7]))
        """
        assert isinstance(data, dict)
        if not hasattr(plot, "add_centroids_x"):
            raise ValueError("Cannot add centroids to this plot")
        if "y0" not in data:
            data["y0"] = np.zeros_like(data["x"], dtype=np.int8)
        self.check_data(data, ("x", "y0", "y1"))
        source = ColumnDataSource(data)
        plot.add_centroids_x(source, **kwargs)

    def add_centroids_y(self, plot: PlotSpectrum, data: ty.Dict, **kwargs):
        """Add horizontal centroids/lines to a particular plot

        Parameters
        ----------
        plot : PlotSpectrum
            plot object to add the band to
        data : dict
            dictionary containing appropriate plot fields
            in this case:
                x0 = the start of the x-coordinates for each horizontal line
                x1 = the end of the x-coordinates for each horizontal
                y = the y-coordinate position for each line
        kwargs :
            dictionary containing plot parameters e.g. line width, line color, transparency, etc...
            must be valid Bokeh fields

        Examples
        --------
        >>> import numpy as np
        >>> x = np.arange(10)
        >>> y = np.arange(10)
        >>> store = PlotStore("")
        >>> _, _, plot = store.plot_spectrum("plot", dict(x=x, y=y))
        >>> store.add_centroids_y(plot, dict(x0=[0, 0, 0], x1=[1, 2, 3], y=[3, 5, 7]))
        """
        if not hasattr(plot, "add_centroids_y"):
            raise ValueError("Cannot add centroids to this plot")
        self.check_data(data, ("y", "x0", "x1"))
        source = ColumnDataSource(data)
        plot.add_centroids_y(source, **kwargs)

    def add_scatter(self, plot: PlotSpectrum, data: ty.Dict, **kwargs):
        """Add scatter points to a particular plot

        Parameters
        ----------
        plot : PlotSpectrum
            plot object to add the band to
        data : dict
            dictionary containing appropriate plot fields
            in this case:
                x = the start of the x-coordinates for each horizontal line
                y = the y-coordinate position for each line
        kwargs :
            dictionary containing plot parameters e.g. line width, line color, transparency, etc...
            must be valid Bokeh fields

        Examples
        --------
        >>> import numpy as np
        >>> x = np.arange(10)
        >>> y = np.arange(10)
        >>> store = PlotStore("")
        >>> _, _, plot = store.plot_spectrum("plot", dict(x=x, y=y))
        >>> store.add_scatter(plot, dict(x0=[0, 0, 0], x1=[1, 2, 3], y=[3, 5, 7]))
        """
        if not hasattr(plot, "add_scatter"):
            raise ValueError("Cannot add scatter points to this plot")
        self.check_data(data, ("x", "y"))
        source = ColumnDataSource(data)
        plot.add_scatter(source, **kwargs)

    @staticmethod
    def link_plots(plot_one: Plot, plot_two: Plot, x_axis: bool = False, y_axis: bool = False):
        """Link the x/y-axis of two plots"""
        if not x_axis and not y_axis:
            return
        if x_axis:
            plot_two.figure.x_range = plot_one.figure.x_range
        if y_axis:
            plot_one.figure.y_range = plot_two.figure.y_range



"""Test imimspy.visualise.store.py"""
import os

import numpy as np
import pytest

try:
    from bokeh.models import Tabs
except ImportError:
    from bokeh.models.widgets import Tabs

from plotski.rgb import ImageRGBA
from plotski.store import PlotStore, containers
from plotski.store.custom import CustomPlotStore


@pytest.fixture
def make_store(tmpdir):
    """Make store."""

    def _store():
        output_dir = os.path.join(tmpdir, "plot-store")
        os.mkdir(output_dir)
        return PlotStore(output_dir)

    return _store


@pytest.fixture
def make_custom_store(tmpdir):
    """Make store."""

    def _store():
        output_dir = os.path.join(tmpdir, "plot-store")
        os.mkdir(output_dir)
        return CustomPlotStore(output_dir)

    return _store


class TestPlotStore:
    @staticmethod
    def test_get_unique_name(make_store):
        store = make_store()
        tab_name = "TEST"
        store.add_tab(tab_name)
        item_name = store.get_unique_name(tab_name)
        assert item_name not in store.tabs

    @staticmethod
    def test_add_tab(make_store):
        store = make_store()
        tab_name = "TEST"
        store.add_tab(tab_name)
        assert tab_name in store.tabs
        tab_name = "TEST2"
        store.add_tab(tab_name)
        assert tab_name in store.tabs
        assert len(store.tabs) == 2

    @staticmethod
    def test_add_tabs(make_store):
        store = make_store()
        tab_names = ["tab 1", "tab 2"]
        store.add_tabs(tab_names)
        assert all([tab_name in store.tabs for tab_name in tab_names])
        assert len(store.tabs) == 2

    @staticmethod
    def test_add_tab_twice(make_store):
        store = make_store()
        tab_name = "TEST"
        store.add_tab(tab_name)
        with pytest.raises(ValueError):
            store.add_tab(tab_name)

    @staticmethod
    def test_add_tab_twice_reset(make_store):
        store = make_store()
        tab_name = "TEST"
        store.add_tab(tab_name)
        store.add_tab(tab_name, override=True)
        assert len(store.tabs) == 1

    @staticmethod
    def test_add_row(make_store):
        store = make_store()
        tab_name = "TEST"
        store.add_tab(tab_name)
        assert tab_name in store.tabs
        row_name = store.add_row(tab_name)
        assert row_name.startswith("row")
        assert row_name in store.tabs[tab_name]

    @staticmethod
    def test_add_row_forgot_tab(make_store):
        store = make_store()
        with pytest.raises(AssertionError):
            _ = store.add_row("TEST")

    @staticmethod
    def test_add_row_plots(make_store):
        store = make_store()
        image = np.random.randint(0, 100, (10, 10))
        tab_name = "heatmaps"
        store.add_tab(tab_name)
        row_name = store.add_row(tab_name)
        store.plot_image(tab_name, dict(image=image), layout_name=row_name)
        store.plot_image(tab_name, dict(image=[image]), layout_name=row_name)
        assert row_name in store.tabs[tab_name]
        assert len(store.tabs[tab_name][row_name]) == 2
        assert type(store[tab_name][row_name]) == containers.Row

    @staticmethod
    def test_add_col(make_store):
        store = make_store()
        tab_name = "TEST"
        store.add_tab(tab_name)
        assert tab_name in store.tabs
        col_name = store.add_col(tab_name)
        assert col_name.startswith("col")
        assert col_name in store.tabs[tab_name]

    @staticmethod
    def test_add_col_forgot_tab(make_store):
        store = make_store()
        with pytest.raises(AssertionError):
            _ = store.add_col("TEST")

    @staticmethod
    def test_add_col_plots(make_store):
        image = np.random.randint(0, 100, (10, 10))
        store = make_store()
        tab_name = "heatmaps"
        store.add_tab(tab_name)
        col_name = store.add_col(tab_name)
        store.plot_image(tab_name, dict(image=[image]), layout_name=col_name)
        store.plot_image(tab_name, dict(image=[image]), layout_name=col_name)
        assert col_name in store.tabs[tab_name]
        assert len(store.tabs[tab_name][col_name]) == 2
        assert type(store[tab_name][col_name]) == containers.Column

    @staticmethod
    def test_add_grid(make_store):
        store = make_store()
        tab_name = "TEST"
        store.add_tab(tab_name)
        assert tab_name in store.tabs
        col_name = store.add_grid(tab_name)
        assert col_name.startswith("grid")
        assert col_name in store.tabs[tab_name]

    @staticmethod
    def test_add_grid_forgot_tab(make_store):
        store = make_store()
        with pytest.raises(AssertionError):
            _ = store.add_grid("TEST")

    @staticmethod
    def test_add_grid_plots(make_store):
        store = make_store()
        image = np.random.randint(0, 100, (10, 10))
        tab_name = "heatmaps"
        store.add_tab(tab_name)
        col_name = store.add_grid(tab_name)
        store.plot_image(tab_name, dict(image=[image]), layout_name=col_name)
        store.plot_image(tab_name, dict(image=[image]), layout_name=col_name)
        store.plot_image(tab_name, dict(image=[image]), layout_name=col_name)
        assert col_name in store.tabs[tab_name]
        assert len(store.tabs[tab_name][col_name]) == 3
        assert type(store[tab_name][col_name]) == containers.Grid

    @staticmethod
    def test_add_spectrum(make_store):
        store = make_store()
        x, y = np.arange(0, 10), np.arange(0, 10)
        tab_name = "spectrum"
        store.add_tab(tab_name)
        store.plot_spectrum(tab_name, dict(x=x, y=y))
        assert "item #0" in store.tabs[tab_name]
        store.plot_spectrum(tab_name, dict(x=x, y=y))
        assert "item #1" in store.tabs[tab_name]

    @staticmethod
    def test_add_scatter(make_store):
        store = make_store()
        x, y = np.arange(0, 10), np.arange(0, 10)
        tab_name = "scatter"
        store.add_tab(tab_name)
        store.plot_scatter(tab_name, dict(x=x, y=y))
        assert "item #0" in store.tabs[tab_name]
        store.plot_scatter(tab_name, dict(x=x, y=y))
        assert "item #1" in store.tabs[tab_name]

    @staticmethod
    def test_add_image(make_store):
        store = make_store()
        image = np.random.randint(0, 100, (10, 10))
        tab_name = "heatmap"
        store.plot_image(tab_name, dict(image=[image]))
        assert "item #0" in store.tabs[tab_name]

    @staticmethod
    def test_add_rgba_image(make_store):
        store = make_store()
        image = [np.random.randint(0, 100, (10, 10))]
        rgba = ImageRGBA(image)
        rgba_img = rgba.rgb
        rgba_intensity = rgba.intensities
        tab_name = "rgba"
        store.plot_rgb_image(tab_name, dict(image=rgba_img))
        store.plot_rgb_image(tab_name, dict(image=[rgba_img], intensities=[rgba_intensity]))
        assert "item #0" in store.tabs[tab_name]

    @staticmethod
    def test_add_annotations_line(make_store):
        store = make_store()
        x, y = np.arange(0, 10), np.arange(0, 10)
        tab_name = "plot"
        store.add_tab(tab_name)
        _, _, plot = store.plot_spectrum(tab_name, dict(x=x, y=y))
        assert "item #0" in store.tabs[tab_name]
        # band annotation
        store.add_band(plot, dict(base=x, lower=y - 3, upper=y + 3))
        # span
        store.add_span(plot, dict(location=1, dimension="width"))
        store.add_span(plot, dict(location=1, dimension="height"))
        store.add_span(plot, dict(location=[5, 3], dimension="height"))
        # add box
        store.add_box(plot, dict(bottom=1, right=3, top=4, left=2))
        with pytest.raises(ValueError):
            store.add_box(plot, dict(bottom=3, right=7))
            store.add_box(plot, dict(top=4, left=2))
        # add labels
        store.add_labels(plot, dict(x=[3, 4], y=[3, 4], text=["label 1", "label 2"]))
        # add centroids
        store.add_segments(plot, dict(x0=[0, 9], x1=[9, 0], y0=[0, 0], y1=[10, 10]))
        store.add_centroids_x(plot, dict(x=[1, 2, 3], y0=[0, 0, 0], y1=[3, 5, 7]))
        store.add_centroids_y(plot, dict(x0=[0, 0, 0], x1=[1, 2, 3], y=[3, 5, 7]))

    @staticmethod
    def test_add_annotations_image(make_store):
        store = make_store()
        image = np.random.randint(0, 100, (10, 10))
        tab_name = "heatmap"
        _, _, plot = store.plot_image(tab_name, dict(image=[image]))
        assert "item #0" in store.tabs[tab_name]
        # band annotation
        store.add_span(plot, dict(location=1, dimension="width"))
        store.add_span(plot, dict(location=1, dimension="height"))
        store.add_span(plot, dict(location=[5, 3], dimension="height"))
        # add box
        store.add_box(plot, dict(bottom=1, right=3, top=4, left=2))
        with pytest.raises(ValueError):
            store.add_box(plot, dict(bottom=3, right=7))
            store.add_box(plot, dict(top=4, left=2))
        # add labels
        store.add_labels(plot, dict(x=[3, 4], y=[3, 4], text=["label 1", "label 2"]))

    @staticmethod
    def test_add_annotations_fail(make_store):
        store = make_store()
        x, y = np.arange(0, 10), np.arange(0, 10)
        image = np.random.randint(0, 100, (10, 10))
        tab_name = "heatmap"
        _, _, plot = store.plot_image(tab_name, dict(image=[image]))
        assert "item #0" in store.tabs[tab_name]
        with pytest.raises(ValueError):
            store.add_band(plot, dict(base=x, lower=y - 3, upper=y + 3))
            # add centroids
            store.add_segments(plot, dict(x0=[0, 9], x1=[9, 0], y0=[0, 0], y1=[10, 10]))
            store.add_centroids_x(plot, dict(x=[1, 2, 3], y0=[0, 0, 0], y1=[3, 5, 7]))
            store.add_centroids_y(plot, dict(x0=[0, 0, 0], x1=[1, 2, 3], y=[3, 5, 7]))

    @staticmethod
    def test_get_layout(make_store):
        store = make_store()
        image = np.random.randint(0, 100, (10, 10))
        tab_name = "heatmap"
        # test individual layout
        store.plot_image(tab_name, dict(image=[image]))
        # test row layout
        row_name = store.add_row(tab_name)
        store.plot_image(tab_name, dict(image=[image]), layout_name=row_name)
        # test column layout
        col_name = store.add_col(tab_name)
        store.plot_image(tab_name, dict(image=[image]), layout_name=col_name)
        # test grid_layout
        grid_name = store.add_grid(tab_name, 2)
        store.plot_image(tab_name, dict(image=[image]), layout_name=grid_name)
        tabs = store.get_layout()
        assert isinstance(tabs, Tabs)

    @staticmethod
    def test_save(make_store):
        store = make_store()
        image = np.random.randint(0, 100, (10, 10))
        tab_name = "heatmap"
        store.add_tab(tab_name)
        store.plot_image(tab_name, dict(image=[image]))

        filepath = os.path.join(store.output_dir, store.filename)
        store.save(show=False)
        assert os.path.exists(filepath)


class TestCustomPlotStore:
    @staticmethod
    def test_add_multiline_spectrum(make_custom_store):
        store = make_custom_store()
        x, y = np.arange(0, 10), np.arange(0, 10)
        tab_name = "spectrum"
        store.add_tab(tab_name)
        store.plot_multiline_spectrum("spectrum", dict(xs=[x, x], ys=[y, y[::-1]]))
        assert "item #0" in store.tabs[tab_name]

    @staticmethod
    def test_add_mass_spectrum(make_custom_store):
        store = make_custom_store()
        x, y = np.arange(0, 10), np.arange(0, 10)
        tab_name = "mass spectrum"
        store.add_tab(tab_name)
        _, _, plot = store.plot_mass_spectrum(tab_name, dict(x=x, y=y))
        assert "item #0" in store.tabs[tab_name]
        store.add_scatter(plot, {"x": x, "y": y})
        store.add_line_plot(plot, {"x": x, "y": y})

    @staticmethod
    def test_add_centroid_mass_spectrum(make_custom_store):
        store = make_custom_store()
        x, y = np.arange(0, 10), np.arange(0, 10)
        tab_name = "mass spectrum"
        store.add_tab(tab_name)
        store.plot_centroid_mass_spectrum(tab_name, dict(x=x, y0=np.zeros_like(y), y1=y))
        assert "item #0" in store.tabs[tab_name]

    @staticmethod
    def test_add_butterfly_mass_spectrum(make_custom_store):
        store = make_custom_store()
        x, y = np.arange(0, 10), np.arange(0, 10)
        tab_name = "butterfly-mass-spectrum"
        store.add_tab(tab_name)
        store.plot_butterfly_mass_spectrum(tab_name, dict(x_top=x, y_top=y, x_bottom=x, y_bottom=-y))
        assert "item #0" in store.tabs[tab_name]
        store.plot_butterfly_mass_spectrum(tab_name, dict(x_top=x, y_top=y, x_bottom=x, y_bottom=-y), add_legend=True)
        assert "item #0" in store.tabs[tab_name]

    @staticmethod
    def test_add_mobilogram(make_custom_store):
        store = make_custom_store()
        x, y = np.arange(0, 10), np.arange(0, 10)
        tab_name = "mass spectrum"
        store.add_tab(tab_name)
        store.plot_mobilogram(tab_name, dict(x=x, y=y))
        assert "item #0" in store.tabs[tab_name]

    @staticmethod
    def test_add_butterfly_mobilogram(make_custom_store):
        store = make_custom_store()
        x, y = np.arange(0, 10), np.arange(0, 10)
        tab_name = "butterfly-mobilogram"
        store.add_tab(tab_name)
        store.plot_butterfly_mobilogram(tab_name, dict(x_top=x, y_top=y, x_bottom=x, y_bottom=-y))
        assert "item #0" in store.tabs[tab_name]



"""Various utilities"""
import random
import typing as ty
from uuid import uuid4

import matplotlib.colors as colors
import numpy as np
from bokeh.models.mappers import LinearColorMapper


def get_colormap(cmap: str):
    """Get matplotlib colormap."""
    try:
        import matplotlib as mpl

        return mpl.colormaps[cmap]
    except AttributeError:
        from matplotlib import cm

        return cm.get_cmap(cmap)


def convert_colormap_to_mapper(array, colormap="viridis", palette=None, z_min=None, z_max=None):
    """Convert matplotlib colormap to Bokeh color mapper

    Parameters
    ----------
    array : np.ndarray
        array
    colormap : str
        name of the colormap
    palette :
    z_min : float
        starting intensity for the colormap
    z_max : float
        final intensity for the colormap

    Returns
    -------
    _palette : Palette
        Bokeh palette
    _color_mapper : LinearColorMapper
        Bokeh colormapper
    """
    array = np.nan_to_num(array)
    if z_min is None:
        z_min = np.round(np.min(array), 2)
    if z_max is None:
        z_max = np.round(np.max(array), 2)

    if palette is None:
        _colormap = get_colormap(colormap)
        _palette = [colors.rgb2hex(m) for m in _colormap(np.arange(_colormap.N))]
    else:
        _palette = palette

    _color_mapper = LinearColorMapper(palette=_palette, low=z_min, high=z_max)

    return _palette, _color_mapper


def rescale(values: ty.Union[np.ndarray, ty.List], new_min: float, new_max: float, dtype=None) -> np.ndarray:
    """Rescale values from one range to another

    Parameters
    ----------
    values : Union[np.ndarray, List]
        input range
    new_min : float
        new minimum value
    new_max : float
        new maximum value
    dtype :
        data type

    Returns
    -------
    new_values : np.ndarray
        rescaled range
    """
    values = np.asarray(values)
    if dtype is None:
        dtype = values.dtype
    old_min, old_max = np.min(values), np.max(values)
    new_values = ((values - old_min) / (old_max - old_min)) * (new_max - new_min) + new_min
    return new_values.astype(dtype)


def convert_hex_to_rgb_1(hex_str, decimals=3):
    """Convert hex color to rgb in range 0-1."""
    hex_color = hex_str.lstrip("#")
    n = len(hex_color)
    rgb = tuple(int(hex_color[i : i + int(n / 3)], 16) for i in range(0, int(n), int(n / 3)))
    return [np.round(rgb[0] / 255.0, decimals), np.round(rgb[1] / 255.0, decimals), np.round(rgb[2] / 255.0, decimals)]


def convert_hex_to_rgb_255(hex_str):
    """Convert hex color to rgb in range 0-255."""
    hex_color = hex_str.lstrip("#")
    n = len(hex_color)
    rgb = list(int(hex_color[i : i + int(n / 3)], 16) for i in range(0, int(n), int(n / 3)))
    return rgb


def get_random_hex_color():
    """Return random hex color"""
    return "#%06x" % random.randint(0, 0xFFFFFF)


def get_min_max(values):
    """Get the minimum and maximum value of an array"""
    return [np.min(values), np.max(values)]


def get_unique_str():
    """Gives random, unique name"""
    return str(uuid4().hex)


def check_key(source, key):
    """Helper function to check source has a particular field"""
    if key in source.data:
        return True
    return False


def check_source(source, keys):
    """Helper function to check source has all of the required fields"""
    missing = []
    for key in keys:
        if key not in source.data:
            missing.append(key)
    if missing:
        raise ValueError(f"Missing '{', '.join(missing)}' from the ColumnDataSource")


def calculate_aspect_ratio(shape, plot_width):
    """Calculate aspect ratio by computing the ratio between the height and width of an array and multiplying it
    by plot width

    Parameters
    ----------
    shape : tuple
        array shape (width, height)
    plot_width : int
        plot width

    Returns
    -------
    plot_width : int
        plot width (not changed)
    plot_height : int
        plot height with array shape being taken into account

    Raises
    ------
    ValueError
        raises ValueError if the shape has fewer than two elements
    """
    if isinstance(shape, np.ndarray):
        shape = shape.shape

    if len(shape) == 2:
        height, width = shape
    else:
        height, width, _ = shape
    if len(shape) < 2:
        raise ValueError("In order to calculate the aspect ratio of the plot, the shape must have two elements (h, w)")

    plot_height = int((height / width) * plot_width)
    return plot_height, plot_width



"""Classes to generate bokeh plots"""
import os
import typing as ty
import webbrowser
from collections.abc import Iterable

import numpy as np
from bokeh.io.export import get_layout_html
from bokeh.layouts import column, row
from bokeh.models import Band, BoxAnnotation, ColumnDataSource, Div, Glyph, LabelSet, Span

from plotski.enums import Position
from plotski.utilities import check_source, get_min_max


class Plot:
    """Base class for all other plots"""

    _div_title, _div_header, _div_footer = None, None, None

    # Data attributes
    DATA_KEYS: ty.Tuple[str, ...] = ()
    TOOLS: ty.Tuple[str, ...] = ("pan", "box_zoom", "reset")
    ACTIVE_DRAG: ty.Optional[str] = None

    # Defaults
    WIDTH = 600
    HEIGHT = 600

    def __init__(
        self,
        output_dir: str,
        source: ColumnDataSource,
        x_axis_label: str = "x",
        y_axis_label: str = "y",
        plot_type: str = "plot",
        initialize: bool = True,
        **kwargs,
    ):
        """Base class for interactive plots

        Note: Not intended to be used directly and instead, please use one of the subclasses which implement the
        actual visualisation(s)

        Parameters
        ----------
        output_dir : str
            Output directory of individual plot.
        source : ColumnDataSource
            Source containing plot data.
        x_axis_label : str
            Label used along x-axis label.
        y_axis_label : str
            Label used along the y-axis label.
        kwargs :
            Dictionary containing key:value parameters to prettify Bokeh plots.
        """
        self.name = None
        self.output_dir = output_dir
        self.plot_type = plot_type
        self.plots: ty.Dict[str, Glyph] = dict()
        self.annotations = dict()
        self._layout = None
        self.div_title = kwargs.pop("title", "")
        self.div_header = kwargs.pop("header", "")
        self._div_header_pos: Position = kwargs.pop("header_pos", Position.ABOVE)
        self.div_footer = kwargs.pop("footer", "")
        self._x_extents: ty.List[float] = []
        self._y_extents: ty.List[float] = []

        # plot attributes
        self.kwargs = kwargs
        self.metadata = dict(x_axis_label=x_axis_label, y_axis_label=y_axis_label)

        self.source = source
        self.check_data_source()

        # initialize options
        self.initialize_options()

        # initialize figure
        self.figure = self.get_figure()
        self.plot()

        # set plot layout and misc data
        if initialize:
            self.set_ranges(**kwargs)
        self.set_hover()
        self.set_figure_attributes()
        self.set_options()
        self.set_figure_dimensions()
        self.set_layout()

    def set_figure_dimensions(self):
        """Set figure dimensions."""
        self.figure.width = self.kwargs.get("width", self.WIDTH)
        self.figure.height = self.kwargs.get("height", self.HEIGHT)

    def check_data_source(self):
        """Ensure that each field in the data source is correct"""
        check_source(self.source, self.DATA_KEYS)

    def set_options(self):
        """Set options."""

    def set_hover(self):
        """Set hover on figure."""

    def initialize_options(self):
        """Initialize extra options"""
        if "tools" not in self.kwargs:
            self.kwargs["tools"] = self.TOOLS
        if "active_drag" not in self.kwargs and self.ACTIVE_DRAG is not None:
            self.kwargs["active_drag"] = self.ACTIVE_DRAG

    def set_figure_attributes(self):
        """Initialize common plot attributes"""
        self.figure.xaxis.axis_label_text_baseline = "bottom"
        self.figure.xaxis.axis_label = self.metadata["x_axis_label"]
        self.figure.yaxis.axis_label = self.metadata["y_axis_label"]

    def set_title(self, text: str, bold: bool = True):
        """Set title text on the title div"""
        if bold:
            text = f"<b>{text}</b>"
        self.div_title.text = text

    @property
    def div_title(self):
        """Return title"""
        return Div(text=f"<b>{self._div_title}</b>")

    @div_title.setter
    def div_title(self, value: str):
        """Set title"""
        self._div_title = value

    @property
    def div_header(self):
        """Return header"""
        return Div(text=self._div_header)

    @div_header.setter
    def div_header(self, value: str):
        """Set header that is associated with a plot."""
        self._div_header = value

    @property
    def div_footer(self):
        """Return footer"""
        return Div(text=self._div_footer, visible=True if self._div_footer else False)

    @div_footer.setter
    def div_footer(self, value: str):
        """Set title"""
        self._div_footer = value

    @property
    def x_axis_label(self) -> str:
        """Get x-axis label."""
        return self.metadata["x_axis_label"]

    @property
    def y_axis_label(self) -> str:
        """Get x-axis label."""
        return self.metadata["y_axis_label"]

    @property
    def layout(self):
        """Get layout"""
        return self.set_layout()

    def set_ranges(self, **kwargs):
        """Set x/y-axis range"""

    def set_layout(self, init_range: bool = True):
        """Setup plot layout

        Each plot consists of three elements:
        TITLE DIV - title of the plot
        FIGURE - Bokeh Plot instance
        ANNOTATION DIV - any comments/annotations provided for particular plot

        --- to be added in the future ---
        TOOLS - interactive tools added for improved visualisation and control purposes
        """
        layout = [self.div_title]
        if self._div_header_pos == Position.ABOVE:
            layout.extend([self.div_header, self.figure])
        elif self._div_header_pos == Position.LEFT:
            layout.append(row(self.div_header, self.figure))
        elif self._div_header_pos == Position.RIGHT:
            layout.append(row(self.div_header, self.figure))
        else:
            layout.extend([self.div_header, self.figure])
        layout.append(self.div_footer)
        self._layout = column(*layout)
        if init_range:
            self.set_ranges()
        return self._layout

    def link_axes(self, x_range=None, y_range=None):
        """Link x- and/or y-axis ranges to another plot.

        It would appear as you can only link plots during creation and not after they've been rendered.
        """
        # if x_range:
        #     if hasattr(x_range, "figure"):
        #         x_range.figure.x_range = self.figure.x_range
        #         # x_range = x_range.figure.x_range
        #     # self.figure.x_range = x_range
        # if y_range:
        #     if hasattr(x_range, "figure"):
        #         y_range = x_range.figure.y_range
        #     self.figure.y_range = y_range

    def add_extents(self, x: ty.Optional[np.ndarray] = None, y: ty.Optional[np.ndarray] = None):
        """Add x-axis extents"""
        if x is not None:
            self._x_extents.append(get_min_max(x))
        if y is not None:
            self._y_extents.append(get_min_max(y))

    def get_extents(self, **kwargs):
        """Get x and y-axis extents"""
        x_min, x_max = get_min_max(self._x_extents)
        x_min, x_max = kwargs.get("x_min", x_min), kwargs.get("x_max", x_max)
        y_min, y_max = get_min_max(self._y_extents)
        y_min, y_max = kwargs.get("y_min", y_min), kwargs.get("y_max", y_max)
        return x_min, x_max, y_min, y_max

    def add_box(self, data: ty.Dict, **kwargs):
        """Add box annotation to the plot"""
        box = BoxAnnotation(**data, **kwargs)
        self.figure.add_layout(box)
        self.annotations[box.id] = (data, "BoxAnnotation")

    def add_patch(self, data: ty.Dict, **kwargs):
        """Add generic polygon/patch to the plot"""
        patch = self.figure.patch(*data, **kwargs)
        self.annotations[patch.id] = (data, "Patch")

    def add_labels(self, source: ColumnDataSource, **kwargs):
        """Add multiple labels to the plot"""
        labels = LabelSet(x="x", y="y", text="text", source=source, **kwargs)
        self.figure.add_layout(labels)
        self.annotations[labels.id] = (source, "LabelSet")

    def add_band(self, source: ColumnDataSource, **kwargs):
        """Add band to the plot"""
        band = Band(
            base="base", lower="lower", upper="upper", source=source, level=kwargs.get("level", "underlay"), **kwargs
        )
        self.figure.add_layout(band)
        self.annotations[band.id] = (source, "Band")

    def add_span(self, data: ty.Dict, **kwargs):
        """Add span to the plot"""
        location = data["location"]
        if not isinstance(location, Iterable):
            location = [location]

        for loc in location:
            span = Span(location=loc, dimension=data["dimension"], **kwargs)
            self.figure.add_layout(span)
            self.annotations[span.id] = (dict(location=loc, dimension=data["dimension"]), "Span")

    def save(self, filepath: ty.Optional[str] = None, show: bool = True):
        """Save Bokeh plot as HTML file

        Parameters
        ----------
        filepath : str
            path where to save the HTML document
        show : bool
            if 'True', newly generated document will be shown in the browser
        """

        if filepath is None:
            filepath = os.path.join(self.output_dir, self.plot_type + ".html")

        html_str = get_layout_html(self.layout)
        with open(filepath, "wb") as f_ptr:
            f_ptr.write(html_str.encode("utf-8"))

        # open figure in browser
        if show:
            webbrowser.open_new_tab(filepath)

    def plot(self):
        """Main plot method."""
        raise NotImplementedError("Must implement method")

    def get_figure(self):
        """Main figure creation method."""
        raise NotImplementedError("Must implement method")



"""Top-level package for plotski."""
from plotski.base import Plot  # noqa
from plotski.store import PlotStore  # noqa
from plotski.store.custom import CustomPlotStore  # noqa

__author__ = "Lukasz G. Migas"
__email__ = "l.g.migas@tudelft.nl"
__version__ = "0.1.0"



"""Container classes."""
import math
import typing as ty
from abc import abstractmethod

T = ty.TypeVar("T")


class Container(ty.MutableSequence):
    """Container for figures contained in the PlotStore using `grid` layout."""

    def __init__(self, data: ty.Iterable = ()):
        self._data = []
        self._data.extend(data)

    @ty.overload
    @abstractmethod
    def __getitem__(self, i: int) -> T:
        ...

    @ty.overload
    @abstractmethod
    def __getitem__(self, s: slice) -> ty.MutableSequence[T]:  # noqa: F811
        ...

    def __getitem__(self, i: int) -> T:  # noqa: F811
        return self._data[i]

    @ty.overload
    @abstractmethod
    def __setitem__(self, i: int, o: T) -> None:
        ...

    @ty.overload
    @abstractmethod
    def __setitem__(self, s: slice, o: ty.Iterable[T]) -> None:  # noqa: F811
        ...

    def __setitem__(self, i: int, o: T) -> None:  # noqa: F811
        self._data[i] = o

    @ty.overload
    @abstractmethod
    def __delitem__(self, i: int) -> None:
        ...

    @ty.overload
    @abstractmethod
    def __delitem__(self, i: slice) -> None:  # noqa: F811
        ...

    def __delitem__(self, i: int) -> None:  # noqa: F811
        del self._data[i]

    def __len__(self) -> int:
        return len(self._data)

    def insert(self, index: int, value: T) -> None:
        """Insert data."""
        self._data.insert(index, value)


class Individual(Container):
    """Container for figures contained in the PlotStore without layout."""


class Row(Container):
    """Container for figures contained in the PlotStore using `row` layout."""


class Column(Container):
    """Container for figures contained in the PlotStore using `column` layout."""


class Grid(Container):
    """Container for figures contained in the PlotStore using `grid` layout."""

    _n_cols: ty.Optional[int] = None

    @property
    def n_cols(self):
        """Number of columns used to represent the grid layout."""
        if self._n_cols is None:
            return math.ceil(math.sqrt(len(self)))
        return self._n_cols

    @n_cols.setter
    def n_cols(self, value: int):
        self._n_cols = value



"""Scatter plot."""
from bokeh.models import ColumnDataSource, HoverTool, Range1d
from bokeh.plotting import figure

from plotski.base import Plot


class PlotScatter(Plot):
    """Scatter plot."""

    # Data parameters
    DATA_KEYS = ("x", "y")
    TOOLS = ("pan, xpan, xbox_zoom, box_zoom, crosshair, reset",)
    ACTIVE_DRAG = "xbox_zoom"

    # Defaults
    WIDTH = 800
    HEIGHT = 400

    def __init__(
        self,
        output_dir: str,
        source: ColumnDataSource,
        x_axis_label: str = "x",
        y_axis_label: str = "y",
        title: str = "Scatter",
        plot_type: str = "Scatter",
        initialize: bool = True,
        **kwargs,
    ):
        Plot.__init__(
            self,
            output_dir,
            source,
            x_axis_label,
            y_axis_label,
            title=title,
            plot_type=plot_type,
            initialize=initialize,
            **kwargs,
        )

    def plot(self):
        """Generate main plot."""
        label = self.kwargs.get("label", "")
        scatter = self.figure.scatter(
            x="x",
            y="y",
            source=self.source,
            name=self.plot_type,
        )
        if label:
            scatter.legend_label = label
        self.plots[scatter.id] = scatter

    def get_figure(self):
        """Get figure."""
        return figure(tools=self.kwargs["tools"], active_drag=self.kwargs["active_drag"])

    def set_hover(self):
        """Set hover."""
        self.figure.add_tools(
            HoverTool(
                show_arrow=True,
                tooltips=[(f"{self.x_axis_label}", "@x"), (f"{self.y_axis_label}", "@y")],
                mode="vline",
                # names=[self.plot_type],
            )
        )

    def set_ranges(self, **kwargs):
        """Set x/y-axis ranges."""
        src = self.source.data
        x_range = self.kwargs.get("x_range", (min(src["x"]), max(src["x"]) * 1.05))
        y_range = self.kwargs.get("y_range", (min(src["y"]), max(src["y"]) * 1.05))
        self.figure.x_range = Range1d(*x_range)
        self.figure.y_range = Range1d(*y_range)
